(***********************************************************************
 * File: Indistinguishability.v
 * $Revision: 408 $
 * $Date: 2011-01-03 13:23:41 -0200 (lun 03 de ene de 2011) $
 * $LastChangedBy: cluna $
 ***********************************************************************
 * Hypervisor_indistinguishability 
 **********************************************************************)


Require Import Exec_invariant.
Require Import Write_Isolation.
Require Import Isolation.
Require Import Traces.
Require Import Hypervisor_imp_hcalls.
Require Import Actions.
Require Import State_equivalence.
Require Import Execution_isolation.
 
Section Indistinguishability.

Variable osi : os_ident.
Variable ctxt : context.

CoInductive same_os_actions : state_stream -> state_stream -> Prop :=
  Os_action: forall (str1 str2:state_stream)
    (a:action) (s1 s2:state),
    same_os_actions str1 str2 -> 
    osi_action osi s1 a ->
    osi_action osi s2 a ->
    same_os_actions (Cons (Ex_State s1 a) str1) (Cons (Ex_State s2 a) str2)
  | Not_os_action1: forall (str1 str2:state_stream)
    (a:action) (s1 s2:state),
    same_os_actions str1 str2 -> 
    ~(osi_action osi s1 a) ->
    same_os_actions (Cons (Ex_State s1 a) str1) str2
  | Not_os_action2: forall (str1 str2:state_stream)
    (a:action) (s1 s2:state),
    same_os_actions str1 str2 -> 
    ~(osi_action osi s2 a) ->
    same_os_actions str1 (Cons (Ex_State s2 a) str2). 


CoInductive osi_equivalent_traces : state_stream -> state_stream -> Prop :=
  eq_trace_osi_action: forall (str1 str2:state_stream)
    (a1 a2:action) (s1 s2:state),
    osi_equivalent_traces str1 str2 -> 
    osi_action osi s1 a1 ->
    osi_action osi s2 a2 ->
    osi_equivalent ctxt s1 s2 osi ->
    osi_equivalent_traces (Cons (Ex_State s1 a1) str1) (Cons (Ex_State s2 a2) str2)
  | eq_trace_not_osi_action1: forall (str1 str2:state_stream)
    (a:action) (s1 s2:state),
    osi_equivalent_traces str1 str2 -> 
    ~(osi_action osi s1 a) ->
    (* osi_equivalent ctxt s1 (ex_state (hd str1)) osi -> *)
    osi_equivalent_traces (Cons (Ex_State s1 a) str1) str2
  | eq_trace_not_osi_action2: forall (str1 str2:state_stream)
    (a:action) (s1 s2:state),
    osi_equivalent_traces str1 str2 -> 
    ~(osi_action osi s2 a) ->
    (* osi_equivalent ctxt s1 (ex_state (hd str2)) osi -> *)
    osi_equivalent_traces str1 (Cons (Ex_State s2 a) str2). 


Theorem Indistinguishability : forall (str1 str2:state_stream),
  run ctxt str1 -> run ctxt str2 ->
  same_os_actions str1 str2 -> 
  osi_equivalent ctxt (ex_state (hd str1)) (ex_state (hd str2)) osi ->
  osi_equivalent_traces str1 str2.
Proof.
cofix str1.
  intros str2 str3 H_run2 H_run3 H_R H_first_osi_eq.
  inversion_clear H_R in H_run2 H_run3 H_first_osi_eq .  
 (* osi is the active_os in s1 and s2 *)
  simpl in H_first_osi_eq.
  constructor; auto.
  unfold run, init_state in H_run2, H_run3.
  simpl in H_run2, H_run3; destruct H_run2; destruct H_run3. 
  apply str1; auto.
  unfold run, init_state.
  inversion_clear H3.
  unfold system_transition in H7; simpl in H7; split; auto.
  apply valid_state_invariant with s1 a; auto.
  inversion H5; auto.
  unfold run, init_state; split; auto.
  apply valid_state_invariant with s2 a; auto.
  generalize full_isolation_proof; unfold full_isolation; intro H_full_isolation_proof.
  apply H_full_isolation_proof with s1 s2 a; auto.
  inversion H3.
  unfold system_transition in H9; simpl in H9; auto.
  inversion H5.
  unfold system_transition in H9; simpl in H9; auto.
(* a is not an osi action wrt to s1 *)
 constructor; auto.
  apply str1; auto.
  apply (run_inv ctxt (Cons (Ex_State s1 a) str0)); auto.
  apply osi_equivalent_trans with s1.
  apply osi_equivalent_symm.
  inversion H_run2.
  unfold init_state in H1; simpl in H1.
  apply not_osi_action_state_eq with a; auto.
  inversion H2; auto.
  simpl in H_first_osi_eq; auto.
(* a is not an osi action wrt to s2 *)
  constructor; auto.
  apply str1; auto.
  apply (run_inv ctxt (Cons (Ex_State s2 a) str4)); auto.
  apply osi_equivalent_trans with s2; auto.
  inversion H_run3.
  unfold init_state in H1; simpl in H1.
  apply not_osi_action_state_eq with a; auto.
  inversion H2; auto.
Qed.

End Indistinguishability.
