(***********************************************************************
 * File: Isolation.v
 * $Revision: 455 $
 * $Date: 2011-01-25 20:29:37 -0200 (Tue, 25 Jan 2011) $
 * $LastChangedBy: jdcampo $
 ***********************************************************************
 * This files proves full os isolation property on valid executions. 
 **********************************************************************)

Require Import State_valid.
Require Import Env.
Require Import Exec.
Require Import Exec_invariant.
Require Export Observers.
Require Import Write_Isolation.
Require Import State_equivalence.
Section Full_OS_Isolation.

Variable ctxt : context.
 

Lemma write_os_isolation_hyper : forall (s1 s2 s1' s2':state) (va:vadd) (val:value),
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  (pre_write va val ctxt s1 \/ pre_write_hyper va val s1) ->
  active_equivalent ctxt s1 s2 ->
  post_all_write va val s1 s1' -> post_all_write va val s2 s2' ->
  osi_hyper_mapping_eq s1' s2' (st_active_os s1).
Proof.
  intros s1 s2 s1' s2' va val H_s1_valid H_s2_valid H_pre1 H_active_eq.
  generalize H_active_eq.
  unfold active_equivalent, aos_equivalent_hyper_mapping,
    osi_hyper_mapping_eq, osi_equivalent_pt_pages, padd_page, vadd_page;
    intros (H_active_os, (H_activity_aos, (H_oss,
      ((H_hyp1_eq, H_hyp2_eq), (H_pt1_eq, H_pt2_eq)))),
      H_post1, H_post2).

  (* Post all write implies modifying a page in the system memory. *)
  (* For s1 *)
  unfold post_all_write in H_post1.
  elim H_post1.
  intros ma1 H_va_mapped1 [newmem1 H_update1 H_state1_assign].
  elim H_state1_assign with System_Memory;
    try (intros [H_abs H']; contradiction H_abs; auto).
  intros [H_smem1 H_interp1].
  unfold state_component_interp in H_interp1.
  unfold update_rw_page in H_update1.
  destruct H_update1 as [pg1 H_pg1_map H_eq_newmem1].
  subst newmem1.
  rewrite H_interp1.

  (* For s2 *) 
  unfold post_all_write in H_post2.
  elim H_post2.
  intros ma2 H_va_mapped2 [newmem2 H_update2 H_state2_assign].
  elim H_state2_assign with System_Memory;
    try (intros [H_abs H']; contradiction H_abs; auto).
  intros [H_smem2 H_interp2].
  unfold state_component_interp in H_interp2.
  unfold update_rw_page in H_update2.
  destruct H_update2 as [pg2 H_pg2_map H_eq_newmem2].
  subst newmem2.
  rewrite H_interp2.

  (* The other state components do not change *)
  generalize (write_affects_system_memory va val s1 s1' H_post1).
  intro H_eqcomp1.
  Simp_Inv_ThreeComps H_eqcomp1 H_eqaos1 H_eqact1 H_eqoss1
    System_Memory Active_OS OS_Activity Guest_OSs.
  Simp_Inv_OneComp H_eqcomp1 H_eqhyp1 System_Memory Hypervisor_SC.
  rewrite <- H_eqhyp1.
  generalize (write_affects_system_memory va val s2 s2' H_post2).
  intro H_eqcomp2.
  Simp_Inv_ThreeComps H_eqcomp2 H_eqaos2 H_eqact2 H_eqoss2
    System_Memory Active_OS OS_Activity Guest_OSs.
  Simp_Inv_OneComp H_eqcomp2 H_eqhyp2 System_Memory Hypervisor_SC.
  rewrite <- H_eqhyp2.

  generalize H_s1_valid; Intros_Valid_Hyp.
  unfold va_mapped_to_ma in H_va_mapped1.
  destruct H_va_mapped1 as [pt_addr1 H_os_curr_map1 [mp1 H_va_mapping1 H_ma1_map]].
  destruct H_os_curr_map1 as [aos1 H_aos1_map [p2m1 H_hyp_map1 H_pt1_map]].
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s1) aos1 p2m1
    as [pt_addr' H_pt'_map [H_pt_type H_pt_owner]]; auto.
  assert (pt_addr' = pt_addr1).
  rewrite <- H_pt'_map in H_pt1_map.
  injection H_pt1_map; auto.
  subst pt_addr'.
  unfold maddress_owner in H_pt_owner.
  assert (exists pt_page,
    Value _ pt_page = map_apply madd_eq (st_memory s1) pt_addr1).
  induction (map_apply madd_eq (st_memory s1) pt_addr1); try discriminate.
  exists v; reflexivity.
  destruct H as [pt_page1 H_pt_page1_map'].
  unfold valid_virtual_mapping in H_valid_virtmap.
  elim H_valid_virtmap with (st_active_os s1) pt_addr1 mp1 pt_page1 va ma1; auto.
  intros pg1' H_pg1'_map [H_acc_owner1 H_notacc_owner1].
  unfold os_accessible in H_acc_owner1, H_notacc_owner1.
  rewrite <- H_pg1'_map in H_pg1_map.
  injection H_pg1_map; intros H_pg1_eq_pg.
  subst pg1'.

  generalize H_s2_valid.
  intros
  (H_trstos_nothypcall_s2,
   (H_running_nothypcall_s2,
    (H_valid_exmd_s2,
     (H_valid_osexmd_s2,
      (H_valid_untosexmd_s2, (H_valid_hyper_s2, (H_valid_virtmap_s2, (H_valid_currpg_s2,
        (H_hyp_inj_s2, H_va_has_pa_s2))))))))).
  unfold va_mapped_to_ma in H_va_mapped2.
  destruct H_va_mapped2 as [pt_addr2 H_os_curr_map2 [mp2 H_va_mapping2 H_ma2_map]].
  destruct H_os_curr_map2 as [aos2 H_aos2_map [p2m2 H_hyp_map2 H_pt2_map]].
  unfold valid_current_page in H_valid_currpg_s2.
  destruct H_valid_currpg_s2 with (st_active_os s2) aos2 p2m2
    as [pt_addr'' H_pt''_map [H_pt''_type H_pt''_owner]]; auto.
  assert (pt_addr'' = pt_addr2).
  rewrite <- H_pt''_map in H_pt2_map.
  injection H_pt2_map; auto.
  subst pt_addr''.
  unfold maddress_owner in H_pt''_owner.
  assert (exists pt_page,
    Value _ pt_page = map_apply madd_eq (st_memory s2) pt_addr2).
  induction (map_apply madd_eq (st_memory s2) pt_addr2); try discriminate.
  exists v; reflexivity.
  destruct H as [pt_page2 H_pt2_map2].
  unfold valid_virtual_mapping in H_valid_virtmap_s2.
  elim H_valid_virtmap_s2 with (st_active_os s2) pt_addr2 mp2 pt_page2 va ma2; auto.
  intros pg H_pg2_map2 [H_acc_owner2 H_notacc_owner2].
  unfold os_accessible in H_acc_owner2, H_notacc_owner2.
  
  rewrite <- H_pg2_map2 in H_pg2_map.
  injection H_pg2_map; intros H_pg1_eq_pg.
  subst pg.

  unfold hypervisor_mapping in H_hyp_map1, H_hyp_map2.

  Focus 2.
  unfold vadd_mapping in H_va_mapping2.
  destruct H_va_mapping2 as [pg' H_pg'_map H_pg'_content].
  rewrite <- H_pg'_map in H_pt2_map2.
  injection H_pt2_map2.
  intros; subst pg'.
  auto.

  Focus 2.
  induction (map_apply madd_eq (st_memory s2) pt_addr2).
  injection H_pt2_map2; intros; subst pt_page2; auto.
  discriminate.

  Focus 2.
  unfold vadd_mapping in H_va_mapping1.
  destruct H_va_mapping1 as [pg' H_pg'_map H_pg'_content].
  rewrite <- H_pg'_map in H_pt_page1_map'.
  injection H_pt_page1_map'.
  intros; subst pg'.
  auto.

  Focus 2.
  induction (map_apply madd_eq (st_memory s1) pt_addr1).
  injection H_pt_page1_map'; intros; subst pt_page1; auto.
  discriminate.

  (* Prove pg1 and pg2 owner are the same *)
  assert (page_owned_by pg1 = page_owned_by pg2) as H_pg1_eq_pg2_own.
  induction (ctxt_vadd_accessible ctxt va).
  rewrite H_acc_owner2; auto.
  rewrite <- H_active_os.
  rewrite H_acc_owner1; auto.
  rewrite H_notacc_owner2; auto.

  (* Prove aos1 = aos2 *)
  assert (aos1 = aos2).
  rewrite <- H_oss in H_aos2_map.
  rewrite <- H_aos2_map in H_aos1_map.
  injection H_aos1_map; auto.
  subst aos2.
 
  intros pa pg H_pa_pg [val' H_pg_content].
  rewrite <- H_pg1_eq_pg2_own.
  destruct H_pa_pg as [ma' H_ma'_mapping H_pg_map]. 

  destruct H_pre1 as [H_pre1 | H_pre1_hyper].
  (* OS Write *)
  unfold pre_write in H_pre1.
  destruct H_pre1 as [H_os_running [H_va_acc H_pre1']].
  unfold va_mapped_to_ma in H_pre1'.
  destruct H_pre1' as [ma1' [pt_addr1' H_os_curr1' [mp1' H_mp1'_map H_ma1'_map]] H_ma1'_ptype].
  assert (ma1'=ma1).
  assert (pt_addr1'=pt_addr1).
  unfold os_current_vadd_mapping_madd in H_os_curr1'.
  destruct H_os_curr1' as [aos1' H_aos1'_map [p2m1'' H_p2m1''_map H_hyp_p2m1'']].
  assert (p2m1''=p2m1).
  rewrite <- H_p2m1''_map in H_hyp_map1.
  injection H_hyp_map1; auto.
  subst p2m1''.
  assert (aos1'=aos1).
  rewrite <- H_aos1'_map in H_aos1_map.
  injection H_aos1_map; auto.
  subst aos1'.
  rewrite <- H_hyp_p2m1'' in H_pt'_map.
  injection H_pt'_map; auto.
  subst pt_addr1'.
  assert (mp1'=mp1).
  unfold vadd_mapping in H_mp1'_map.
  destruct H_mp1'_map as [pt_page1' H_pt_page1'_map H_mp1'_map].
  assert (pt_page1'=pt_page1).
  rewrite <- H_pt_page1'_map in H_pt_page1_map'.
  injection H_pt_page1_map'; auto.
  subst pt_page1'.
  unfold vadd_mapping in H_va_mapping1.
  destruct H_va_mapping1 as [pt_page1' H_pt_page1'_map' H_mp1'_map'].
  assert (pt_page1'=pt_page1).
  rewrite <- H_pt_page1'_map' in H_pt_page1_map'.
  injection H_pt_page1_map'; auto.
  subst pt_page1'.
  rewrite <- H_mp1'_map in H_mp1'_map'.
  injection H_mp1'_map'; auto.
  subst mp1'.
  rewrite <- H_ma1'_map in H_ma1_map.
  injection H_ma1_map; auto.
  subst ma1'.

  unfold va_has_valid_pa in H_va_has_pa.
  elim H_va_has_pa with (st_active_os s1) (os_current_page aos1) pt_addr1 pt_page1 mp1 va ma1; auto.
  intros pa1 H_pa1_map.

  (* Prove PT_c mp1 = page_content pt_page1 *)
  assert (PT_c mp1 = page_content pt_page1) as H_pt_page1_content.
  unfold vadd_mapping in H_va_mapping1.
  destruct H_va_mapping1 as [pt_page1' H_pt_page1'_map H_pt_page1'_content].
  assert (pt_page1'=pt_page1).
  rewrite <- H_pt_page1'_map in H_pt_page1_map'.
  injection H_pt_page1_map'; auto.
  subst pt_page1'.
  assumption.

  case (padd_eq pa1 pa).
  (* pa1 = pa *)
  intros H_pa1_eq_pa; subst pa1.

  assert (ma'=ma1).
  unfold hypervisor_mapping_madd in H_pa1_map, H_ma'_mapping.
  destruct H_pa1_map as [p2m1' H_p2m1'_map H_ma1_map'].
  destruct H_ma'_mapping as [p2m1'' H_p2m1''_map H_ma1_map''].
  assert (p2m1'=p2m1'').
  rewrite <- H_p2m1''_map in H_p2m1'_map.
  injection H_p2m1'_map; auto.
  subst p2m1''.
  rewrite <- H_ma1_map'' in H_ma1_map'.
  injection H_ma1_map'; auto.
  subst ma'.
  rewrite lemma_map_add_apply in H_pg_map; auto.

  elim H_pt1_eq with (os_current_page aos1) pt_page1 mp1; auto.
  intros pt_page2' [pt_addr2' H_hyp_pt_addr2' H_pt_page2_map]
    [mp2' H_pg'_content H_virt_map_eq].
  elim H_virt_map_eq with va pa ma1 pg1; auto.
  intros ma2' H_ma2'_map [H_hyp_ma2' H_pg_map2].
  assert (ma2'=ma2).
  assert (pt_addr2' = pt_addr2).
  unfold hypervisor_mapping_madd in H_hyp_pt_addr2'.
  destruct H_hyp_pt_addr2' as [p2m2' H_p2m2'_map H_pt_addr2'_map].
  assert (p2m2'=p2m2).
  rewrite H_active_os in H_p2m2'_map.
  rewrite <- H_p2m2'_map in H_hyp_map2.
  injection H_hyp_map2; auto.
  subst p2m2'.
  rewrite <- H_pt_addr2'_map in H_pt2_map.
  injection H_pt2_map; auto.
  subst pt_addr2'.
  assert (pt_page2' = pt_page2).
  rewrite <- H_pt_page2_map in H_pt2_map2.
  injection H_pt2_map2; auto.
  subst pt_page2'.
  assert (mp2' = mp2).
  unfold vadd_mapping in H_va_mapping2.
  destruct H_va_mapping2 as [pt_page2' H_pt_page2'_map H_pt_page2'_content'].
  assert (pt_page2' = pt_page2).
  rewrite <- H_pt_page2'_map in H_pt2_map2.
  injection H_pt2_map2; auto.
  subst pt_page2'.
  rewrite <- H_pg'_content in H_pt_page2'_content'.
  injection H_pt_page2'_content'; auto.
  subst mp2'.
  rewrite <- H_ma2'_map in H_ma2_map.
  injection H_ma2_map; auto.
  subst ma2'.

  exists ma2; auto.
  rewrite lemma_map_add_apply; auto.

  unfold hypervisor_mapping_madd.
  exists pt_addr1; auto.
  exists p2m1; auto.
  
  (* pa1 <> pa *)
  intros H_pa1_neq_pa.
  assert (ma1 <> ma').
  intros H_ma1_eq_ma'; apply H_pa1_neq_pa.

  unfold injective_hyper_mappings in H_injective_hyper.
  rewrite H_injective_hyper with (st_active_os s1) p2m1 pa pa1; auto.

  unfold hypervisor_mapping_madd in H_ma'_mapping, H_pa1_map.
  destruct H_ma'_mapping as [p2m1' H_p2m1'_map H_ma'_map].
  destruct H_pa1_map as [p2m1'' H_p2m1''_map H_ma''_map].

  assert (p2m1'=p2m1).
  rewrite <- H_p2m1'_map in H_hyp_map1.
  injection H_hyp_map1; auto.
  subst p2m1'.
  assert (p2m1''=p2m1).
  rewrite <- H_p2m1''_map in H_hyp_map1.
  injection H_hyp_map1; auto.
  subst p2m1''.
  subst ma'.
  rewrite <- H_ma''_map; auto.

  rewrite lemma_map_add_apply_distinct in H_pg_map; auto.

  elim H_hyp1_eq with pa pg; auto.
  intros ma2' H_ma2'_map H_pg1''_map.
  
  exists ma2'; auto.
  
  elim H_pt1_eq with (os_current_page aos1) pt_page1 mp1; auto.
  intros pt_page2' [pt_addr2' H_hyp_pt_addr2' H_pt_page2_map]
    [mp2' H_pg'_content H_virt_map_eq].
  elim H_virt_map_eq with va pa1 ma1 pg1; auto.
  intros ma'' H_ma''_map [H_hyp_ma'' H_pg_map2].
  assert (ma2<>ma2').
  intros H_ma2_eq_ma2'.
  subst ma2'.
  apply H_pa1_neq_pa.

  unfold injective_hyper_mappings in H_hyp_inj_s2.
  rewrite H_hyp_inj_s2 with (st_active_os s1) p2m2 pa1 pa; auto.

  rewrite H_active_os; auto.
  unfold hypervisor_mapping_madd in H_hyp_ma'', H_ma2'_map.
  destruct H_hyp_ma'' as [p2m2' H_p2m2'_map H_ma'_map].
  destruct H_ma2'_map as [p2m2'' H_p2m2''_map H_ma''_map'].
  assert (p2m2' = p2m2).
  rewrite H_active_os in H_p2m2'_map.
  rewrite <- H_p2m2'_map in H_hyp_map2.
  injection H_hyp_map2; auto.
  subst p2m2'.
  assert (p2m2'' = p2m2).
  rewrite H_active_os in H_p2m2''_map.
  rewrite <- H_p2m2''_map in H_hyp_map2.
  injection H_hyp_map2; auto.
  subst p2m2''.
  rewrite <- H_ma'_map.
  rewrite <- H_ma''_map'.

  assert (pt_addr2'=pt_addr2).
  unfold hypervisor_mapping_madd in H_hyp_pt_addr2'.
  destruct H_hyp_pt_addr2' as [p2m2' H_p2m2'_map' H_pt_addr2'_map H_pt_page2'_map].
  assert (p2m2'=p2m2).
  rewrite <- H_p2m2'_map' in H_p2m2''_map.
  injection H_p2m2''_map; auto.
  subst p2m2'.
  rewrite <- H_pt_addr2'_map in H_pt2_map.
  injection H_pt2_map; auto.
  subst pt_addr2'.
  assert (pt_page2'=pt_page2).
  rewrite <- H_pt_page2_map in H_pt2_map2.
  injection H_pt2_map2; auto.
  subst pt_page2'.
  unfold vadd_mapping in H_va_mapping2.
  destruct H_va_mapping2 as [pt_page2' H_pt_page2'_map' H_pt_page2'_content].
  assert (pt_page2'=pt_page2).
  rewrite <- H_pt_page2'_map' in H_pt_page2_map.
  injection H_pt_page2_map; auto.
  subst pt_page2'.
  assert (mp2'=mp2).
  rewrite <- H_pt_page2'_content in H_pg'_content.
  injection H_pg'_content; auto.
  subst mp2'.
  rewrite <- H_ma''_map in H_ma2_map.
  injection H_ma2_map.
  intros H_ma2_eq_ma''; symmetry in H_ma2_eq_ma''.
  subst ma''; reflexivity.

  rewrite lemma_map_add_apply_distinct; auto.

  unfold os_current_vadd_mapping_madd in H_os_curr1'.
  destruct H_os_curr1' as [aos1' H_aos1'_map [p2m1' H_hyp_map1' H_pt_addr1'_map]].
  assert (p2m1'=p2m1).
  rewrite <- H_hyp_map1' in H_hyp_map1.
  injection H_hyp_map1; auto.
  subst p2m1'.
  assert (aos1' = aos1).
  rewrite <- H_aos1'_map in H_aos1_map.
  injection H_aos1_map; auto.
  subst aos1'.
  unfold hypervisor_mapping_madd.

  unfold padd_page.
  exists pt_addr1; auto.

  unfold hypervisor_mapping_madd.
  exists p2m1; auto.

  unfold os_current_vadd_mapping_madd in H_os_curr1'.
  destruct H_os_curr1' as [aos1' H_aos1'_map [p2m1' H_hyp_map1' H_pt_addr1'_map]].
  assert (p2m1'=p2m1).
  rewrite <- H_hyp_map1' in H_hyp_map1.
  injection H_hyp_map1; auto.
  subst p2m1'.
  assert (aos1' = aos1).
  rewrite <- H_aos1'_map in H_aos1_map.
  injection H_aos1_map; auto.
  subst aos1'.
  unfold hypervisor_mapping_madd.
  unfold padd_page.
  exists ma'; auto.
  exists val'; auto.
  exists p2m1; auto.

  (* Prove PT_c mp1 = page_content pt_page1 *)
  unfold vadd_mapping in H_va_mapping1.
  destruct H_va_mapping1 as [pt_page1' H_pt_page1'_map H_pt_page1'_content].
  assert (pt_page1'=pt_page1).
  rewrite <- H_pt_page1'_map in H_pt_page1_map'.
  injection H_pt_page1_map'; auto.
  subst pt_page1'.
  assumption.

  (* Hyper Write *)
  assert (page_owned_by pg1 = Hyp_PO) as H_pg1_owner_hyp.
  unfold pre_write_hyper in H_pre1_hyper.
  destruct H_pre1_hyper as [H_hyp_running [ma1' H_ma1'_map [H_ma1'_ptype H_ma1'_owner]]].
  unfold maddress_owner in H_ma1'_owner.
  assert (ma1'=ma1).
  unfold va_mapped_to_ma in H_ma1'_map.
  destruct H_ma1'_map as [pt_addr1' H_os_curr1 [mp1' H_mp1'_map H_ma1'_map]].
  unfold vadd_mapping in H_mp1'_map.
  destruct H_mp1'_map as [pt_page1' H_pt_page1'_map H_pt_page1'_content].
  unfold os_current_vadd_mapping_madd in H_os_curr1.
  destruct H_os_curr1 as [aos1' H_aos1'_map [p2m1' H_hyp_map1' H_pt_addr1'_map]].
  assert (p2m1'=p2m1).
  rewrite <- H_hyp_map1' in H_hyp_map1.
  injection H_hyp_map1; auto.
  subst p2m1'.
  assert (aos1' = aos1).
  rewrite <- H_aos1'_map in H_aos1_map.
  injection H_aos1_map; auto.
  subst aos1'.
  assert (pt_addr1'=pt_addr1).
  rewrite <- H_pt_addr1'_map in H_pt'_map.
  injection H_pt'_map; auto.
  subst pt_addr1'.
  assert (pt_page1'=pt_page1).
  rewrite <- H_pt_page1'_map in H_pt_page1_map'.
  injection H_pt_page1_map'; auto.
  subst pt_page1'.
  unfold vadd_mapping in H_va_mapping1.
  destruct H_va_mapping1 as [pt_page1' H_pt_page1'_map' H_pt_page1'_content'].
  assert (pt_page1'=pt_page1).
  rewrite <- H_pt_page1'_map' in H_pt_page1'_map.
  injection H_pt_page1'_map; auto.
  subst pt_page1'.
  assert (mp1'=mp1).
  rewrite <- H_pt_page1'_content' in H_pt_page1'_content.
  injection H_pt_page1'_content; auto.
  subst mp1'.
  rewrite <- H_ma1'_map in H_ma1_map.
  injection H_ma1_map; auto.
  subst ma1'.
  induction (map_apply madd_eq (st_memory s1) ma1); try discriminate.
  assert (pg1=v).
  injection H_pg1'_map; auto.
  subst v.
  assumption.

  assert (ma1 <> ma').
  intros H_ma1_eq_ma'.
  subst ma'.
  rewrite lemma_map_add_apply in H_pg_map.

  assert (page_owned_by pg1 = Os_PO (st_active_os s1)).
  unfold valid_hypervisor in H_valid_hyper.
  elim H_valid_hyper with (st_active_os s1) pa p2m1.
  intros [ma1' H_ma1'_map H_ma1'_owner].
  assert (ma1'=ma1).
  unfold hypervisor_mapping_madd in H_ma'_mapping.
  destruct H_ma'_mapping as [p2m1' H_p2m1'_map H_ma1_map_padd].
  assert (p2m1'=p2m1).
  rewrite <- H_p2m1'_map in H_hyp_map1.
  injection H_hyp_map1; auto.
  subst p2m1'.
  rewrite <- H_ma1'_map in H_ma1_map_padd.
  injection H_ma1_map_padd; auto.
  subst ma1'.
  unfold maddress_owner in H_ma1'_owner.
  induction (map_apply madd_eq (st_memory s1) ma1); try discriminate.
  injection H_pg1'_map.
  intro H_pg1_eq_v; subst v.
  assumption.

  intros H_abs.
  unfold hypervisor_mapping_madd in H_ma'_mapping.
  destruct H_ma'_mapping as [p2m1' H_p2m1'_map H_ma'_map'].
  assert (p2m1' = p2m1).
  rewrite <- H_p2m1'_map in H_hyp_map1.
  injection H_hyp_map1; auto.
  subst p2m1'.
  rewrite H_abs in H_ma'_map'; discriminate.
  assumption.
  rewrite H in H_pg1_owner_hyp; discriminate.
  rewrite lemma_map_add_apply_distinct in H_pg_map; auto.

  (*  Prove that pa cannot change *)
  elim H_hyp1_eq with pa pg; auto.
  intros ma2' H_ma2'_map H_pg1''_map.
  exists ma2'; auto.
  
  (* Now prove that ma2<>ma2' *)
  assert (ma2<>ma2').
  intro H_ma2_eq_ma2'.
  subst ma2'.
  assert (pg=pg2).
  rewrite <- H_pg1''_map in H_pg2_map2.
  injection H_pg2_map2; auto.
  subst pg.
  rewrite H_pg1_eq_pg2_own in H_pg1_owner_hyp.
  assert (page_owned_by pg2 = Os_PO (st_active_os s1)).
  unfold valid_hypervisor in H_valid_hyper_s2.
  elim H_valid_hyper_s2 with (st_active_os s1) pa p2m2.
  intros [ma2' H_ma2'_map' H_ma2'_owner].
  assert (ma2'=ma2).
  unfold hypervisor_mapping_madd in H_ma2'_map.
  destruct H_ma2'_map as [p2m2' H_p2m2'_map H_ma2_map_padd].
  assert (p2m2'=p2m2).
  rewrite H_active_os in H_p2m2'_map.
  rewrite <- H_p2m2'_map in H_hyp_map2.
  injection H_hyp_map2; auto.
  subst p2m2'.
  rewrite <- H_ma2'_map' in H_ma2_map_padd.
  injection H_ma2_map_padd; auto.
  subst ma2'.
  unfold maddress_owner in H_ma2'_owner.
  induction (map_apply madd_eq (st_memory s2) ma2); try discriminate.
  injection H_pg1''_map.
  intro H_pg1_eq_v; subst v.
  assumption.

  intros H_abs.
  unfold hypervisor_mapping_madd in H_ma2'_map.
  destruct H_ma2'_map as [p2m2' H_p2m2'_map H_ma'_map'].
  assert (p2m2' = p2m2).
  rewrite H_active_os in H_p2m2'_map.
  rewrite <- H_p2m2'_map in H_hyp_map2.
  injection H_hyp_map2; auto.
  subst p2m2'.
  rewrite H_abs in H_ma'_map'; discriminate.
  rewrite <- H_active_os in H_hyp_map2; assumption.
  rewrite H0 in H_pg1_owner_hyp; discriminate.
  rewrite lemma_map_add_apply_distinct; auto.
  exists ma'; auto.

  exists val'; auto.
Qed.

Lemma write_os_isolation_pt_pages : forall (s1 s2 s1' s2':state) (va:vadd) (val:value),
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  ((pre_write va val ctxt s1) /\ (pre_write va val ctxt s2)) \/
  ((pre_write_hyper va val s1) /\ (pre_write_hyper va val s2)) ->
  active_equivalent ctxt s1 s2 ->
  post_all_write va val s1 s1' -> post_all_write va val s2 s2' ->
  osi_pt_pages_eq ctxt s1' s2' (st_active_os s1).
Proof.
  intros s1 s2 s1' s2' va val H_s1_valid H_s2_valid H_pre H_active_eq.
  generalize H_active_eq.
  unfold active_equivalent, aos_equivalent_hyper_mapping,
    aos_equivalent_pt_pages, osi_pt_pages_eq, padd_page, vadd_page;
    intros (H_active_os, (H_activity_aos, (H_oss,
      ((H_hyp1_eq, H_hyp2_eq), (H_pt1_eq, H_pt2_eq)))),
      H_post1, H_post2).

  (* Post all write implies modifying a page in the system memory. *)
  (* For s1 *)
  unfold post_all_write in H_post1.
  elim H_post1.
  intros ma1 H_va_mapped1 [newmem1 H_update1 H_state1_assign].
  elim H_state1_assign with System_Memory;
    try (intros [H_abs H']; contradiction H_abs; auto).
  intros [H_smem1 H_interp1].
  unfold state_component_interp in H_interp1.
  unfold update_rw_page in H_update1.
  destruct H_update1 as [pg H_pg_map1 H_eq_newmem1].
  subst newmem1.
  rewrite H_interp1.

  (* For s2 *) 
  unfold post_all_write in H_post2.
  elim H_post2.
  intros ma2 H_va_mapped2 [newmem2 H_update2 H_state2_assign].
  elim H_state2_assign with System_Memory;
    try (intros [H_abs H']; contradiction H_abs; auto).
  intros [H_smem2 H_interp2].
  unfold state_component_interp in H_interp2.
  unfold update_rw_page in H_update2.
  destruct H_update2 as [pg2 H_pg2_map H_eq_newmem2].
  subst newmem2.
  rewrite H_interp2.

  (* The other state components do not change *)
  generalize (write_affects_system_memory va val s1 s1' H_post1).
  intro H_eqcomp1.
  Simp_Inv_ThreeComps H_eqcomp1 H_eqaos1 H_eqact1 H_eqoss1
    System_Memory Active_OS OS_Activity Guest_OSs.
  Simp_Inv_OneComp H_eqcomp1 H_eqhyp1 System_Memory Hypervisor_SC.
  rewrite <- H_eqhyp1.
  generalize (write_affects_system_memory va val s2 s2' H_post2).
  intro H_eqcomp2.
  Simp_Inv_ThreeComps H_eqcomp2 H_eqaos2 H_eqact2 H_eqoss2
    System_Memory Active_OS OS_Activity Guest_OSs.
  Simp_Inv_OneComp H_eqcomp2 H_eqhyp2 System_Memory Hypervisor_SC.
  rewrite <- H_eqhyp2.

  intros pt_padd pt_page1 mp1 [pt_addr1 H_pt_addr1_map H_pt_page1_map] H_pt_page1_content.

  (* pt_addr1 is not the written page, as pt_addr1 is a PT page *) 
  assert (pt_addr1<>ma1) as H_pt_addr1_neq_ma1.
  intro H_pt_addr1_eq_ma1.
  rewrite H_pt_addr1_eq_ma1 in H_pt_page1_map.
  rewrite lemma_map_add_apply in H_pt_page1_map.
  injection H_pt_page1_map; intro H_pt_page1_value.
  rewrite H_pt_page1_value in H_pt_page1_content.
  discriminate H_pt_page1_content.

  rewrite lemma_map_add_apply_distinct in H_pt_page1_map; auto.
  elim H_pt1_eq with pt_padd pt_page1 mp1; try (exists pt_addr1); auto.
  intros pt_page2 [pt_addr2 H_pt_addr2_map H_pt_page2_map] [mp2 H_pt_page2_content H_virt_map_eq].
  exists pt_page2.
  exists pt_addr2; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma2_eq_pt_addr2.
  destruct H_pre as [[H_pre1 H_pre2] | [H_pre1_hyper H_pre2_hyper]].
  (* OS write *)
  unfold pre_write in H_pre2.
  destruct H_pre2 as [H_os_running [H_va_accessible [ma2' H_ma2'_map H_pg2_type]]].
  assert (ma2'=ma2).
  apply va_mapped_same_ma with s2 va; auto.
  subst ma2'; clear H_ma2'_map.
  unfold page_type in H_pg2_type.
  rewrite <- H_ma2_eq_pt_addr2 in H_pt_page2_map.
  rewrite <- H_pt_page2_map in H_pg2_type.
  rewrite <- H_pt_page2_content in H_pg2_type.
  discriminate H_pg2_type.
  (* Hypervisor write *)
  unfold pre_write_hyper in H_pre2_hyper.
  destruct H_pre2_hyper as [H_hyper_running [ma2' H_ma2'_map [H_pg2_type H_pg2'_owner]]].
  assert (ma2'=ma2).
  apply va_mapped_same_ma with s2 va; auto.
  subst ma2'; clear H_ma2'_map.
  unfold page_type in H_pg2_type.
  rewrite <- H_ma2_eq_pt_addr2 in H_pt_page2_map.
  rewrite <- H_pt_page2_map in H_pg2_type.
  rewrite <- H_pt_page2_content in H_pg2_type.
  discriminate H_pg2_type.
  exists mp2; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.

  (* Get the pa that maps to ma1 *)
  unfold va_mapped_to_ma in H_va_mapped1.
  destruct H_va_mapped1 as [pt_curr1_addr H_os_curr_map1 [curr_mp1 H_va_mapping1 H_ma1_map]].
  destruct H_os_curr_map1 as [aos1 H_aos1_map [p2m1 H_p2m1_map H_pt_curr_addr1_map]].
  unfold vadd_mapping in H_va_mapping1.
  destruct H_va_mapping1 as [pt_curr1 H_pt_curr1_map H_pt_curr1_content].
  unfold hypervisor_mapping in H_p2m1_map.
  unfold va_mapped_to_ma in H_va_mapped2.
  destruct H_va_mapped2 as [pt_curr2_addr H_os_curr_map2 [curr_mp2 H_va_mapping2 H_ma2_map]].
  destruct H_os_curr_map2 as [aos2 H_aos2_map [p2m2 H_p2m2_map H_pt_curr2_addr_map]].
  unfold vadd_mapping in H_va_mapping2.
  destruct H_va_mapping2 as [pt_curr2 H_pt_curr2_map H_pt_curr2_content].
  unfold hypervisor_mapping in H_p2m2_map.
  elim H_pt1_eq with (os_current_page aos1) pt_curr1 curr_mp1; auto.
  2: exists pt_curr1_addr; unfold hypervisor_mapping_madd;
    try exists p2m1; auto.

  assert (aos1=aos2) as H_aos1_eq_aos2.
  rewrite <- H_oss in H_aos2_map.
  rewrite <- H_aos1_map in H_aos2_map.
  injection H_aos2_map; auto.

  intros pt_curr2' [pt_curr2_addr' H_pt_curr2_addr_hyp_map H_pt_curr2'_map].
  unfold hypervisor_mapping_madd in H_pt_curr2_addr_hyp_map.
  destruct H_pt_curr2_addr_hyp_map as [p2m2' H_p2m2'_map H_pt_curr2_addr_hyp_map].
  rewrite <- H_active_os in H_p2m2_map.
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intro; subst p2m2'; clear H_p2m2'_map.
  rewrite H_aos1_eq_aos2 in H_pt_curr2_addr_hyp_map.
  rewrite <- H_pt_curr2_addr_map in H_pt_curr2_addr_hyp_map.
  injection H_pt_curr2_addr_hyp_map; intro; subst pt_curr2_addr'; clear H_pt_curr2_addr_hyp_map.
  rewrite <- H_pt_curr2_map in H_pt_curr2'_map.
  injection H_pt_curr2'_map; intro; subst pt_curr2'; clear H_pt_curr2'_map.

  intros [curr_mp2' H_pt_curr2_content' H_virt_map_eq_va].
  rewrite <- H_pt_curr2_content in H_pt_curr2_content'.
  injection H_pt_curr2_content'; intro; subst curr_mp2'; clear H_pt_curr2_content'.

  destruct H_pre as [[H_pre1 H_pre2] | [H_pre1_hyper H_pre2_hyper]].
  (* OS write *)
  unfold pre_write in H_pre1.
  destruct H_pre1 as [H_os_running [H_va_accessible [ma1' H_ma1'_map H_pg_type]]].
  assert (ma1'=ma1).
  apply va_mapped_same_ma with s1 va; auto.
  unfold va_mapped_to_ma.
  exists pt_curr1_addr.
  unfold os_current_vadd_mapping_madd.
  exists aos1; auto.
  exists p2m1; auto.
  exists curr_mp1; auto.
  unfold vadd_mapping.
  exists pt_curr1; auto.
  subst ma1'; clear H_ma1'_map.
  
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_currpg H_valid_hyper H_injective_hyper.
  unfold va_has_valid_pa in H_va_has_pa.
  elim H_va_has_pa with (st_active_os s1) (os_current_page aos1) pt_curr1_addr pt_curr1 curr_mp1 va ma1; auto.
  intros pa H_pa_map.
  elim H_virt_map_eq_va with va pa ma1 pg; auto.
  intros ma2' H_ma2'_map [H_ma2_hyp_map H_pg_map2].
  rewrite <- H_ma2_map in H_ma2'_map.
  injection H_ma2'_map; intro; subst ma2'; clear H_ma2'_map.

  assert (pg=pg2).
  rewrite <- H_pg2_map in H_pg_map2.
  injection H_pg_map2; auto.
  subst pg2; clear H_pg2_map.

  case (madd_eq ma1 ma').
  (* ma1 is the written address *)
  intro H_ma1_eq_ma'.

  assert (pa=pa') as H_pa_eq_pa'.
  apply hyp_mapped_same_paddr with ctxt s1 (st_active_os s1) ma1; auto.
  rewrite H_ma1_eq_ma'; assumption.
  elim H_virt_map_eq with va' pa' ma' pg; auto.
  intros ma H_ma_map [H_ma_hyp_map H_pg1_map2'].
  assert (ma2=ma) as H_ma2_eq_ma.
  apply hyp_mapped_same_maddr with s2 (st_active_os s2) pa; auto.
  rewrite <- H_active_os; assumption.
  rewrite <- H_active_os.
  rewrite H_pa_eq_pa'; assumption.
  exists ma2; auto.
  rewrite H_ma2_eq_ma; assumption.
  split.
  rewrite H_ma2_eq_ma; assumption.
  rewrite H_ma1_eq_ma' in H_pg'_map.
  rewrite lemma_map_add_apply in H_pg'_map.
  injection H_pg'_map; intro H_pg'_value; clear H_pg'_map.
  rewrite H_pg'_value.
  rewrite lemma_map_add_apply; reflexivity.
  rewrite <- H_ma1_eq_ma'; assumption.
  rewrite <- H_ma1_eq_ma'; assumption.

  (* ma' is not the written page *)
  intro H_ma1_ne_ma'.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_map, H_pg'_type; auto.
  rewrite <- H_pg'_map in H_pg'_type.
  elim H_virt_map_eq with va' pa' ma' pg'; auto.
  intros ma H_ma_map [H_ma_hyp_map H_pg'_map2].
  exists ma; auto.
  split; auto.

  2: unfold page_type; rewrite <- H_pg'_map;
    induction (page_content pg'); try discriminate H_pg'_type; reflexivity.

  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma2_eq_ma.
  rewrite <- H_ma2_eq_ma in H_ma_hyp_map.

  assert (pa=pa') as H_pa_eq_pa'.
  apply hyp_mapped_same_paddr with ctxt s2 (st_active_os s2) ma; auto.
  rewrite <- H_active_os.
  rewrite <- H_ma2_eq_ma; assumption.
  rewrite <- H_active_os.
  rewrite <- H_ma2_eq_ma; assumption.
  apply H_ma1_ne_ma'.
  apply hyp_mapped_same_maddr with s1 (st_active_os s1) pa'; auto.
  rewrite <- H_pa_eq_pa'; assumption.
  exists p2m1; auto.

  (* Hypervisor write *)
  unfold pre_write_hyper in H_pre1_hyper.
  destruct H_pre1_hyper as [H_hyper_running [ma1' H_ma1'_map [H_pg1_type H_pg1_owner]]].

  assert (ma1'=ma1).
  apply va_mapped_same_ma with s1 va; auto.
  unfold va_mapped_to_ma.
  exists pt_curr1_addr.
  unfold os_current_vadd_mapping_madd.
  exists aos1; auto.
  exists p2m1; auto.
  exists curr_mp1; auto.
  unfold vadd_mapping.
  exists pt_curr1; auto.
  subst ma1'; clear H_ma1'_map.

  case (madd_eq ma1 ma').
  (* ma' is the written address *)
  intro H_ma1_eq_ma'.
  rewrite H_ma1_eq_ma' in H_pg1_owner.
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  destruct (H_valid_hyper (st_active_os s1) pa' p2m1 H_p2m1_map)
    as [[ma_1 H_ma_1_map H_ma_1_owner]|H_error].
  assert (ma_1=ma').
  apply hyp_mapped_same_maddr with  s1 (st_active_os s1) pa'; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  subst ma_1.
  rewrite H_pg1_owner in H_ma_1_owner.
  discriminate H_ma_1_owner.
  unfold hypervisor_mapping_madd in H_ma'_hyp_map.
  destruct H_ma'_hyp_map as [p2m1' H_p2m1'_map H_ma'_hyp_map].
  rewrite <- H_p2m1_map in H_p2m1'_map.
  injection H_p2m1'_map; intro; subst p2m1'; clear H_p2m1'_map.
  rewrite <- H_ma'_hyp_map in H_error.
  discriminate H_error.

  (* ma' is not the written page *)
  intro H_ma1_ne_ma'.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_map, H_pg'_type; auto.
  rewrite <- H_pg'_map in H_pg'_type.
  elim H_virt_map_eq with va' pa' ma' pg'; auto.
  intros ma H_ma_map [H_ma_hyp_map H_pg'_map2].
  exists ma; auto.
  split; auto.

  2: unfold page_type; rewrite <- H_pg'_map;
    induction (page_content pg'); try discriminate H_pg'_type; reflexivity.

  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma2_eq_ma.
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  rewrite H_active_os in H_p2m2_map.
  destruct (H_valid_hyper (st_active_os s2) pa' p2m2 H_p2m2_map)
    as [[ma_2 H_ma_2_map H_ma_2_owner]|H_error].
  assert (ma_2=ma).
  apply hyp_mapped_same_maddr with  s2 (st_active_os s2) pa'; auto.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.
  rewrite <- H_active_os; auto.
  subst ma_2.
  rewrite <- H_ma2_eq_ma in H_ma_2_owner.
  destruct H_pre2_hyper as [H_hyper_running2 [ma2' H_ma2'_map [H_pg1'_type H_pg1'_owner]]].
  assert (ma2'=ma2).
  apply va_mapped_same_ma with s2 va; auto.
  unfold va_mapped_to_ma.
  exists pt_curr2_addr.
  unfold os_current_vadd_mapping_madd.
  exists aos2; auto.
  exists p2m2; auto.
  exists curr_mp2; auto.
  unfold vadd_mapping.
  exists pt_curr2; auto.
  subst ma2'.
  rewrite H_ma_2_owner in H_pg1'_owner.
  discriminate H_pg1'_owner.
  unfold hypervisor_mapping_madd in H_ma_hyp_map.
  destruct H_ma_hyp_map as [p2m2' H_p2m2'_map H_ma_hyp_map].
  rewrite <- H_active_os in H_p2m2_map.
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intro; subst p2m2'; clear H_p2m2'_map.
  rewrite <- H_ma_hyp_map in H_error.
  discriminate H_error.
Qed.

Lemma write_os_isolation : forall (s1 s2 s1' s2':state) (va:vadd) (val:value),
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  ((pre_write va val ctxt s1 /\ pre_write va val ctxt s2)) \/
  ((pre_write_hyper va val s1) /\ (pre_write_hyper va val s2)) ->
  active_equivalent ctxt s1 s2 ->
  post_all_write va val s1 s1' -> post_all_write va val s2 s2' ->
  active_equivalent ctxt s1' s2'.
Proof.
  intros s1 s2 s1' s2' va val H_s1_valid H_s2_valid H_pre H_active_eq.
  generalize H_active_eq.
  unfold active_equivalent, padd_page, vadd_page;
    intros (H_active_os, (H_activity_aos, (H_oss,
      ((H_hyp1_eq, H_hyp2_eq), (H_pt1_eq, H_pt2_eq)))),
      H_post1, H_post2).

  unfold aos_equivalent_hyper_mapping, aos_equivalent_pt_pages,
    osi_equivalent_hyper_mapping, osi_equivalent_pt_pages in *.

  (* The other state components do not change *)
  generalize (write_affects_system_memory va val s1 s1' H_post1).
  intro H_eqcomp1.
  Simp_Inv_ThreeComps H_eqcomp1 H_eqaos1 H_eqact1 H_eqoss1
    System_Memory Active_OS OS_Activity Guest_OSs.
  rewrite <- H_eqaos1.
  rewrite <- H_eqact1.
  rewrite <- H_eqoss1.
  generalize (write_affects_system_memory va val s2 s2' H_post2).
  intro H_eqcomp2.
  Simp_Inv_ThreeComps H_eqcomp2 H_eqaos2 H_eqact2 H_eqoss2
    System_Memory Active_OS OS_Activity Guest_OSs.
  rewrite <- H_eqaos2.
  rewrite <- H_eqact2.
  rewrite <- H_eqoss2.

  generalize (active_eq_symm ctxt s1 s2); intro H_s2_equiv_s1.

  repeat split; auto.

  apply write_os_isolation_hyper with s2 va val; auto.
  destruct H_pre as [[H_pre1 H_pre2] | [H_pre1_hyper H_pre2_hyper]]; auto.
  rewrite H_active_os.
  apply write_os_isolation_hyper with s1 va val; auto.
  destruct H_pre as [[H_pre1 H_pre2] | [H_pre1_hyper H_pre2_hyper]]; auto.
  apply write_os_isolation_pt_pages with s2 va val; auto.
  rewrite H_active_os.
  apply write_os_isolation_pt_pages with s1 va val; auto.
  destruct H_pre as [[H_pre1 H_pre2] | [H_pre1_hyper H_pre2_hyper]]; auto.
Qed.

(* New hyper action *)
Lemma new_hyper_os_isolation_hyper : forall (s1 s2 s1' s2':state) (va:vadd) (ma:madd),
  active_equivalent ctxt s1 s2 ->
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  post_new_hyper va ma s1 s1' -> post_new_hyper va ma s2 s2' ->
  osi_hyper_mapping_eq s1' s2' (st_active_os s1).
Proof.
  intros s1 s2 s1' s2' va ma H_active_eq H_s1_valid H_s2_valid.
  generalize H_active_eq.
  unfold active_equivalent;
    intros (H_active_os, (H_activity_aos,
      (H_oss, ((H_hyp1, H_hyp2), (H_page_s1s2, H_page_s2s1)))), H_post1, H_post2).
  generalize H_post1 H_post2.
  intros H_post1' H_post2'.
  unfold post_new_hyper in H_post1', H_post2'.
  destruct H_post1' as [system H_system_map [pt_addr1 H_pt_addr1_map [newmem1 H_update1 H_state1_assign]]].
  destruct H_post2' as [system2 H_system2_map [pt_addr2 H_pt_addr2_map [newmem2 H_update2 H_state2_assign]]].
  rewrite <- H_oss in H_system2_map.
  rewrite <- H_system_map in H_system2_map.
  injection H_system2_map; intro; subst system2; clear H_system2_map.

  unfold state_component_assign, state_component_update
    in H_state1_assign, H_state2_assign.

  unfold osi_hyper_mapping_eq, padd_page.

  (* Only System Memory changes *)
  assert (forall cmp, cmp <> System_Memory ->
    state_component_interp s1 cmp = state_component_interp s1' cmp)
    as H_eqcomp1.
  intros cmp H_eqcomp1; elim H_state1_assign with cmp.
  intros [H_abs H_mem].
  contradiction H_abs; auto.
  intros [H_eq1 H_interp1]; assumption.

  assert (forall cmp, cmp <> System_Memory ->
    state_component_interp s2 cmp = state_component_interp s2' cmp)
    as H_eqcomp2.
  intros cmp H_eqcomp2; elim H_state2_assign with cmp.
  intros [H_abs H_mem].
  contradiction H_abs; auto.
  intros [H_eq1 H_interp2]; assumption.
  Simp_Inv_ThreeComps H_eqcomp1 H_eqaos1 H_eqhyp1 H_eqoss1 System_Memory Active_OS Hypervisor_SC Guest_OSs.
  rewrite <- H_eqhyp1. 
  Simp_Inv_ThreeComps H_eqcomp2 H_eqaos2 H_eqhyp2 H_eqoss2 System_Memory Active_OS Hypervisor_SC Guest_OSs.
  rewrite <- H_eqhyp2; try (intros; discriminate).
  elim H_state1_assign with System_Memory.
  intros [H_triv H_smem1].
  2: intros [H_abs H_smem1]; contradiction H_abs; reflexivity.
  elim H_state2_assign with System_Memory.
  intros [H_triv2 H_smem2].
  2: intros [H_abs H_smem2]; contradiction H_abs; reflexivity.
  clear H_triv H_triv2.
  unfold state_component_interp in H_smem1, H_smem2.
  rewrite H_smem1. rewrite H_smem2.

  unfold update_pt_page in H_update1, H_update2.
  destruct H_update1 as [old_pt1 H_old_pt1_map [old_mp1 H_old_pt1_content H_newmem1]].
  destruct H_update2 as [old_pt2 H_old_pt2_map [old_mp2 H_old_pt2_content H_newmem2]].

  intros pa' pg' [ma' H_ma'_hyp_map H_pg'_map] [val H_pg'_content].
  (* ma' cannot be the changed pt_addr1, as it contains a RW page *)
  assert (ma'<>pt_addr1) as H_ma'_ne_pt_addr1.
  intro H_ma'_eq_pt_addr1.
  rewrite H_newmem1 in H_pg'_map.
  rewrite H_ma'_eq_pt_addr1 in H_pg'_map.
  rewrite lemma_map_add_apply in H_pg'_map.
  injection H_pg'_map; intro H_pg'_value; rewrite H_pg'_value in H_pg'_content; discriminate H_pg'_content.
  
  elim H_hyp1 with pa' pg'; auto.
  intros ma2 H_ma2_map H_pg'_map2.
  exists ma2; auto.

  assert (ma2<>pt_addr2) as H_ma2_ne_pt_addr2.
  intro H_ma2_eq_pt_addr2.
  unfold os_current_vadd_mapping_madd in H_pt_addr2_map.
  destruct H_pt_addr2_map as [system2 H_system2_map [p2m2 H_p2m2_map H_pt_addr2_map]].
  rewrite <- H_oss in H_system2_map.
  rewrite <- H_system_map in H_system2_map.
  injection H_system2_map; intro; subst system2; clear H_system2_map.
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_injective_hyper H_va_has_pa.
  unfold valid_current_page in H_valid_currpg.
  elim H_valid_currpg with (st_active_os s1) system p2m2; auto.
  intros pt_addr2' H_pt_addr2'_map [H_pt_addr2'_type H_pt_addr2'_owner].
  rewrite <- H_pt_addr2_map in H_pt_addr2'_map.
  injection H_pt_addr2'_map; intro; subst pt_addr2'; clear H_pt_addr2'_map.
  rewrite <- H_ma2_eq_pt_addr2 in H_pt_addr2'_type.
  unfold page_type in H_pt_addr2'_type.
  rewrite <- H_pg'_map2 in H_pt_addr2'_type.
  rewrite H_pg'_content in H_pt_addr2'_type.
  discriminate H_pt_addr2'_type.

  rewrite H_active_os.
  rewrite <- H_oss.
  assumption.
  rewrite H_active_os.
  assumption.

  rewrite H_newmem2.
  rewrite lemma_map_add_apply_distinct; auto.
  unfold padd_page.
  exists ma'; auto.
  rewrite H_newmem1 in H_pg'_map.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  exists val; auto.
Qed.

Lemma new_hyper_os_isolation_pt_pages : forall (s1 s2 s1' s2':state) (va:vadd) (ma:madd),
  active_equivalent ctxt s1 s2 ->
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  pre_new_hyper va ma ctxt s1 -> pre_new_hyper va ma ctxt s2 ->
  post_new_hyper va ma s1 s1' -> post_new_hyper va ma s2 s2' ->
  osi_pt_pages_eq ctxt s1' s2' (st_active_os s1).
Proof.
  intros s1 s2 s1' s2' va ma H_active_eq H_s1_valid H_s2_valid H_pre1 H_pre2.
  generalize H_active_eq.
  unfold active_equivalent;
    intros (H_active_os, (H_activity_aos,
      (H_oss, ((H_hyp1, H_hyp2), (H_page_s1s2, H_page_s2s1)))), H_post1, H_post2).
  generalize H_post1 H_post2.
  intros H_post1' H_post2'.
  unfold post_new_hyper in H_post1', H_post2'.
  destruct H_post1' as [system H_system_map [pt_addr1 H_pt_addr1_map [newmem1 H_update1 H_state1_assign]]].
  destruct H_post2' as [system2 H_system2_map [pt_addr2 H_pt_addr2_map [newmem2 H_update2 H_state2_assign]]].

  rewrite <- H_oss in H_system2_map.
  rewrite <- H_system_map in H_system2_map.
  injection H_system2_map; intro; subst system2; clear H_system2_map.

  unfold state_component_assign, state_component_update
    in H_state1_assign, H_state2_assign.

  unfold pre_new_hyper in H_pre1, H_pre2.
  destruct H_pre1 as [H_hyp_running1 [H_ma_ptype1 [H_ma_owner1 H_va_not_accessible1]]].
  destruct H_pre2 as [H_hyp_running2 [H_ma_ptype2 [H_ma_owner2 H_va_not_accessible2]]].

  unfold osi_pt_pages_eq, padd_page.

  (* Only System Memory changes *)
  assert (forall cmp, cmp <> System_Memory ->
    state_component_interp s1 cmp = state_component_interp s1' cmp)
    as H_eqcomp1.
  intros cmp H_eqcomp1; elim H_state1_assign with cmp.
  intros [H_abs H_mem].
  contradiction H_abs; auto.
  intros [H_eq1 H_interp1]; assumption.

  assert (forall cmp, cmp <> System_Memory ->
    state_component_interp s2 cmp = state_component_interp s2' cmp)
    as H_eqcomp2.
  intros cmp H_eqcomp2; elim H_state2_assign with cmp.
  intros [H_abs H_mem].
  contradiction H_abs; auto.
  intros [H_eq1 H_interp2]; assumption.
  Simp_Inv_ThreeComps H_eqcomp1 H_eqaos1 H_eqhyp1 H_eqoss1
    System_Memory Active_OS Hypervisor_SC Guest_OSs.
  rewrite <- H_eqhyp1.
  Simp_Inv_ThreeComps H_eqcomp2 H_eqaos2 H_eqhyp2 H_eqoss2
    System_Memory Active_OS Hypervisor_SC Guest_OSs.
  rewrite <- H_eqhyp2; try (intros; discriminate).
  elim H_state1_assign with System_Memory.
  2: intros [H_abs H_interp]; contradiction H_abs; reflexivity.
  intros [H_triv H_smem1].
  elim H_state2_assign with System_Memory.
  2: intros [H_abs H_interp]; contradiction H_abs; reflexivity.
  intros [H_triv2 H_smem2].
  clear H_triv H_triv2.
  unfold state_component_interp in H_smem1, H_smem2.
  rewrite H_smem1. rewrite H_smem2.

  unfold update_pt_page in H_update1, H_update2.
  destruct H_update1 as [old_pt1 H_old_pt1_map [old_mp1 H_old_pt1_content H_newmem1]].
  destruct H_update2 as [old_pt2 H_old_pt2_map [old_mp2 H_old_pt2_content H_newmem2]].

  intros pt_padd pt_page1' mp1' [pt_addr1' H_pt_addr1'_map H_pt_page1'_map] H_pt_page1'_content.

  unfold hypervisor_mapping_madd in H_pt_addr1'_map.
  destruct H_pt_addr1'_map as [p2m1 H_p2m1_map H_pt_addr1'_map].
  unfold hypervisor_mapping in H_p2m1_map.

  unfold os_current_vadd_mapping_madd in H_pt_addr1_map.
  destruct H_pt_addr1_map as [system1 H_system1_map [p2m1' H_p2m1'_map H_pt_addr1_curr_map]].
  rewrite <- H_system_map in H_system1_map.
  injection H_system1_map; intro; subst system1; clear H_system1_map.
  unfold hypervisor_mapping in H_p2m1'_map.
  rewrite <- H_p2m1_map in H_p2m1'_map.
  injection H_p2m1'_map; intro; subst p2m1'; clear H_p2m1'_map.

  case (madd_eq pt_addr1 pt_addr1').
  intro H_pt_addr1'_eq_pt_addr1.

  assert (pt_padd=(os_current_page system)).
  apply hyp_mapped_same_paddr with ctxt s1 (st_active_os s1) pt_addr1; auto.
  rewrite H_pt_addr1'_eq_pt_addr1.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  subst pt_padd.

  rewrite H_newmem1 in H_pt_page1'_map.
  rewrite H_pt_addr1'_eq_pt_addr1 in H_pt_page1'_map.
  rewrite lemma_map_add_apply in H_pt_page1'_map.
  injection H_pt_page1'_map; intro H_pt_page1'_value.
  
  unfold os_current_vadd_mapping_madd in H_pt_addr2_map.

  destruct H_pt_addr2_map as [system' H_system'_map [p2m2 H_p2m2_map H_pt_addr2_map]].
  unfold hypervisor_mapping in H_p2m2_map.
  rewrite <- H_oss in H_system'_map.
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  
  exists (Page (PT_c (map_add vadd_eq old_mp2 va ma)) (page_owned_by old_pt2)).
  exists pt_addr2.
  unfold hypervisor_mapping_madd.
  rewrite H_active_os.
  exists p2m2; auto.

  rewrite H_newmem2.
  rewrite lemma_map_add_apply.
  reflexivity.

  exists (map_add vadd_eq old_mp2 va ma); auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va_accessible H_pg'_type H_ma'_hyp_map.

  unfold hypervisor_mapping_madd in H_ma'_hyp_map.
  destruct H_ma'_hyp_map as [p2m1' H_p2m1'_map H_ma'_hyp_map].
  rewrite <- H_p2m1_map in H_p2m1'_map.
  injection H_p2m1'_map; intro; subst p2m1'; clear H_p2m1'_map.

  (* va' is not the new address *)
  assert (va<>va') as H_va_ne_va'.
  intro H_va_eq_va'.
  rewrite <- H_va_eq_va' in H_va_accessible.
  apply H_va_not_accessible1; assumption.

  rewrite lemma_map_add_apply_distinct; auto.
  assert (ma<>ma') as H_ma_ne_ma'.
  intro H_ma_eq_ma'.
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_currpg
      H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  elim H_valid_hyper with (st_active_os s1) pa' p2m1; auto.
  intros [ma_1 H_ma_1_map H_ma'_owner].
  rewrite <- H_ma'_hyp_map in H_ma_1_map.
  injection H_ma_1_map; intro; subst ma_1; clear H_ma_1_map.
  rewrite <- H_ma_eq_ma' in H_ma'_owner.
  rewrite H_ma_owner1 in H_ma'_owner.
  discriminate H_ma'_owner.
  intro H_error.
  rewrite <- H_ma'_hyp_map in H_error.
  discriminate H_error.
  
  assert ((os_current_page system) <> pa') as H_pa'_not_curr_pt.
  intro H_pa'_curr_pt.
  assert (ma'=pt_addr1') as H_ma'_eq_pt_addr1'.
  apply hyp_mapped_same_maddr with s1 (st_active_os s1) pa'; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  unfold hypervisor_mapping_madd.
  rewrite <- H_pa'_curr_pt.
  exists p2m1; auto.

  rewrite H_ma'_eq_pt_addr1' in H_pg'_type.
  unfold page_type in H_pg'_type.
  rewrite H_newmem1 in H_pg'_type.
  rewrite H_pt_addr1'_eq_pt_addr1 in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_type.
  simpl in H_pg'_type.
  discriminate H_pg'_type.

  assert (pt_addr1<>ma') as H_pt_addr1_ne_ma'.
  intro H_pt_addr1_eq_ma'.
  apply H_pa'_not_curr_pt.
  apply hyp_mapped_same_paddr with ctxt s1 (st_active_os s1) pt_addr1; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  rewrite H_pt_addr1_eq_ma'.
  assumption.

  elim H_page_s1s2 with (os_current_page system) old_pt1 old_mp1; auto.
  intros pt_page2' H_pt_page2'_map [mp2' H_pt_page2'_content H_virt_map_eq].
  unfold padd_page in H_pt_page2'_map.
  destruct H_pt_page2'_map as [pt_addr2' H_pt_addr2'_map H_pt_page2'_map].
  unfold hypervisor_mapping_madd in H_pt_addr2'_map.
  destruct H_pt_addr2'_map as [p2m2' H_p2m2'_map H_pt_addr2'_map].
  rewrite H_active_os in H_p2m2'_map.
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intro; subst p2m2'; clear H_p2m2'_map.
  rewrite <- H_pt_addr2_map in H_pt_addr2'_map.
  injection H_pt_addr2'_map; intro; subst pt_addr2'; clear H_pt_addr2'_map.

  elim H_virt_map_eq with va' pa' ma' pg'; auto.
  intros ma2 H_ma2_map [H_ma2_hyp_map H_pg'_map2].
  exists ma2.
  rewrite <- H_old_pt2_map in H_pt_page2'_map.
  injection H_pt_page2'_map; intro; subst pt_page2'; clear H_pt_page2'_map.
  rewrite <- H_old_pt2_content in H_pt_page2'_content.
  injection H_pt_page2'_content; intro; subst mp2'; clear H_pt_page2'_content.
  assumption.

  split; auto.
  rewrite H_newmem2.
  rewrite lemma_map_add_apply_distinct; auto.

  intro H_pt_addr2_eq_ma2.
  apply H_pa'_not_curr_pt.
  apply hyp_mapped_same_paddr with ctxt s2 (st_active_os s2) ma2; auto.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.
  rewrite <- H_pt_addr2_eq_ma2.
  assumption.
  rewrite <- H_active_os; assumption.
  
  rewrite H_pt_page1'_value in H_pt_page1'_content.
  simpl in H_pt_page1'_content.
  injection H_pt_page1'_content; intro H_mp1'_value.
  rewrite H_mp1'_value in H_ma'_map.
  rewrite lemma_map_add_apply_distinct in H_ma'_map; auto.

  rewrite H_newmem1 in H_pg'_map.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.

  rewrite H_newmem1 in H_pg'_type.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_type; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  unfold padd_page.
  exists pt_addr1; auto.
  rewrite H_pt_addr1'_eq_pt_addr1.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.

  (* If pt_addr1' is not the current PT page *)
  intro H_pt_addr1'_ne_pt_addr1.
  destruct H_pt_addr2_map as [system' H_system'_map [p2m2 H_p2m2_map H_pt_addr2_map]].
  unfold hypervisor_mapping in H_p2m2_map.
  rewrite <- H_oss in H_system'_map.
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  elim H_page_s1s2 with pt_padd pt_page1' mp1'; auto.
  intros pt_page2 H_pt_page2_map [mp2 H_pt_page2_content H_virt_map_eq].
  unfold padd_page in H_pt_page2_map.
  destruct H_pt_page2_map as [pt_addr2' H_pt_addr2'_map H_pt_page2_map].
  unfold hypervisor_mapping_madd in H_pt_addr2'_map.
  destruct H_pt_addr2'_map as [p2m2' H_p2m2'_map H_pt_addr2'_map].
  rewrite H_active_os in H_p2m2'_map.
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intro; subst p2m2'; clear H_p2m2'_map.
  
  assert (pt_padd<>(os_current_page system)) as H_pt_padd_ne_curr_pt.
  intro H_pt_padd_eq_curr_pt.
  apply H_pt_addr1'_ne_pt_addr1.
  apply hyp_mapped_same_maddr with s1 (st_active_os s1) pt_padd.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  rewrite H_pt_padd_eq_curr_pt.
  assumption.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  
  assert (pt_addr2'<>pt_addr2) as H_pt_addr2'_ne_pt_addr2.
  intro H_pt_addr2'_eq_pt_addr2.
  apply H_pt_padd_ne_curr_pt.
  apply hyp_mapped_same_paddr with ctxt s2 (st_active_os s2) pt_addr2; auto.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.
  rewrite <- H_pt_addr2'_eq_pt_addr2; assumption.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.

  exists pt_page2.
  exists pt_addr2'.
  unfold hypervisor_mapping_madd.
  rewrite H_active_os.
  exists p2m2; auto.
  rewrite H_newmem2.
  rewrite lemma_map_add_apply_distinct; auto.
  exists mp2; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va_accessible H_pg'_type H_ma'_hyp_map.

  rewrite H_newmem1 in H_pg'_map, H_pg'_type.
  unfold page_type in H_pg'_type.

  assert (pt_addr1<>ma') as H_pt_addr1_neq_ma'.
  intro H_pt_addr1_eq_ma'.
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_injective_hyper H_va_has_pa.
  unfold valid_current_page in H_valid_currpg.
  elim H_valid_currpg with (st_active_os s1) system p2m1; auto.
  intros pt_addr1'' H_pt_addr1''_map [H_pt_addr1_type H_pt_addr1_owner].
  rewrite <- H_pt_addr1_curr_map in H_pt_addr1''_map.
  injection H_pt_addr1''_map; intro; subst pt_addr1''; clear H_pt_addr1''_map.
  rewrite H_pt_addr1_eq_ma' in H_pt_addr1_type.
  rewrite H_pt_addr1_eq_ma' in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_type.
  simpl in H_pg'_type.
  discriminate H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_map, H_pg'_type; auto.
  rewrite <- H_pg'_map in H_pg'_type.

  elim H_virt_map_eq with va' pa' ma' pg'; auto.
  intros ma2 H_ma2_map [H_ma2_hyp_map H_pg'_map2].
  exists ma2; auto.
  split; auto.
  
  rewrite H_newmem2.
  assert (pt_addr2<>ma2) as H_pt_addr2_neq_ma2.
  intro H_pt_addr2_eq_ma2.
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_injective_hyper H_va_has_pa.
  unfold valid_current_page in H_valid_currpg.
  elim H_valid_currpg with (st_active_os s2) system p2m2; auto.
  intros pt_addr2'' H_pt_addr2''_map [H_pt_addr2_type H_pt_addr2_owner].
  rewrite <- H_pt_addr2_map in H_pt_addr2''_map.
  injection H_pt_addr2''_map; intro; subst pt_addr2''; clear H_pt_addr2''_map.
  rewrite H_pt_addr2_eq_ma2 in H_pt_addr2_type.
  unfold page_type in H_pt_addr2_type.
  rewrite <- H_pg'_map2 in H_pt_addr2_type.
  rewrite H_pg'_type in H_pt_addr2_type.
  discriminate H_pt_addr2_type.

  rewrite <- H_oss; auto.

  rewrite lemma_map_add_apply_distinct; auto.
  unfold page_type.
  rewrite <- H_pg'_map.
  rewrite H_pg'_type; reflexivity.
  unfold padd_page.
  exists pt_addr1'.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  rewrite H_newmem1 in H_pt_page1'_map.
  rewrite lemma_map_add_apply_distinct in H_pt_page1'_map; auto.
Qed.

Lemma new_hyper_os_isolation : forall (s1 s2 s1' s2':state) (va:vadd) (ma:madd),
  active_equivalent ctxt s1 s2 ->
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  pre_new_hyper va ma ctxt s1 -> pre_new_hyper va ma ctxt s2 -> 
  post_new_hyper va ma s1 s1' -> post_new_hyper va ma s2 s2' ->
  active_equivalent ctxt s1' s2'.
Proof.
  intros s1 s2 s1' s2' va ma H_active_eq H_s1_valid H_s2_valid H_pre1 H_pre2.
  generalize H_active_eq.
  unfold active_equivalent;
    intros (H_active_os, (H_activity_aos,
      (H_oss, ((H_hyp1, H_hyp2), (H_page_s1s2, H_page_s2s1)))), H_post1, H_post2).
  generalize H_post1 H_post2.
  intros H_post1' H_post2'.
  unfold post_all_new in H_post1', H_post2'.
  destruct H_post1' as [system H_system_map [pt_addr1 H_pt_addr1_map [newmem1 H_update1 H_state1_assign]]].
  destruct H_post2' as [system2 H_system2_map [pt_addr2 H_pt_addr2_map [newmem2 H_update2 H_state2_assign]]].

  unfold state_component_assign, state_component_update
    in H_state1_assign, H_state2_assign.

  unfold aos_equivalent_hyper_mapping, aos_equivalent_pt_pages,
    osi_equivalent_hyper_mapping, osi_equivalent_pt_pages.

  (* Only System Memory changes *)
  assert (forall cmp, cmp <> System_Memory ->
    state_component_interp s1 cmp = state_component_interp s1' cmp)
    as H_eqcomp1.
  intros cmp H_eqcomp1; elim H_state1_assign with cmp.
  intros [H_abs H_mem].
  contradiction H_abs; auto.
  intros [H_eq1 H_interp1]; assumption.

  assert (forall cmp, cmp <> System_Memory ->
    state_component_interp s2 cmp = state_component_interp s2' cmp)
    as H_eqcomp2.
  intros cmp H_eqcomp2; elim H_state2_assign with cmp.
  intros [H_abs H_mem].
  contradiction H_abs; auto.
  intros [H_eq1 H_interp2]; assumption.
  Simp_Inv_ThreeComps H_eqcomp1 H_eqaos1 H_eqact1 H_eqoss1
    System_Memory Active_OS OS_Activity Guest_OSs.
  rewrite <- H_eqaos1.
  rewrite <- H_eqact1. 
  rewrite <- H_eqoss1.
  Simp_Inv_ThreeComps H_eqcomp2 H_eqaos2 H_eqact2 H_eqoss2
    System_Memory Active_OS OS_Activity Guest_OSs.
  rewrite <- H_eqaos2. 
  rewrite <- H_eqact2. 
  rewrite <- H_eqoss2. 

  generalize (active_eq_symm ctxt s1 s2); intro H_s2_equiv_s1.
  
  repeat split; try assumption.
  
  apply new_hyper_os_isolation_hyper with s2 va ma; auto.
  rewrite H_active_os.
  apply new_hyper_os_isolation_hyper with s1 va ma; auto.
  apply new_hyper_os_isolation_pt_pages with s2 va ma; auto.
  rewrite H_active_os.
  apply new_hyper_os_isolation_pt_pages with s1 va ma; auto.
Qed.

(* Delete trusted and hyper *)
Lemma del_os_isolation_hyper : forall (s1 s2 s1' s2':state)
  (va:vadd) (pt_addr1 pt_addr2:madd),
  active_equivalent ctxt s1 s2 ->
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  current_vadd_mapping_madd s1 pt_addr1 -> current_vadd_mapping_madd s2 pt_addr2 ->
  post_all_del va pt_addr1 s1 s1' -> post_all_del va pt_addr2 s2 s2' ->
  osi_hyper_mapping_eq s1' s2' (st_active_os s1).
Proof.
  intros s1 s2 s1' s2' va pt_addr1 pt_addr2 H_active_eq H_s1_valid H_s2_valid H_pt_addr1_map H_pt_addr2_map.
  generalize H_active_eq.
  unfold active_equivalent;
    intros (H_active_os, (H_activity_aos,
      (H_oss, ((H_hyp1, H_hyp2), (H_page_s1s2, H_page_s2s1)))), H_post1, H_post2).
  generalize H_post1 H_post2.
  intros H_post1' H_post2'.
  unfold post_all_del in H_post1', H_post2'.
  destruct H_post1' as [newmem1 H_drop1 H_state1_assign].
  destruct H_post2' as [newmem2 H_drop2 H_state2_assign].
  unfold assign_system_memory, state_component_assign, state_component_update in H_state1_assign, H_state2_assign.

  unfold osi_hyper_mapping_eq, padd_page.

  (* Only System Memory changes *)
  assert (forall cmp, cmp <> System_Memory ->
    state_component_interp s1 cmp = state_component_interp s1' cmp)
    as H_eqcomp1.
  intros cmp H_eqcomp1; elim H_state1_assign with cmp.
  intros [H_abs H_mem].
  contradiction H_abs; auto.
  intros [H_eq1 H_interp1]; assumption.
  
  assert (forall cmp, cmp <> System_Memory ->
    state_component_interp s2 cmp = state_component_interp s2' cmp)
    as H_eqcomp2.
  intros cmp H_eqcomp2; elim H_state2_assign with cmp.
  intros [H_abs H_mem].
  contradiction H_abs; auto.
  intros [H_eq1 H_interp2]; assumption.
  Simp_Inv_ThreeComps H_eqcomp1 H_eqaos1 H_eqhyp1 H_eqoss1
    System_Memory Active_OS Hypervisor_SC Guest_OSs.
  rewrite <- H_eqhyp1.
  Simp_Inv_ThreeComps H_eqcomp2 H_eqaos2 H_eqhyp2 H_eqoss2
    System_Memory Active_OS Hypervisor_SC Guest_OSs.
  rewrite <- H_eqhyp2.
  elim H_state1_assign with System_Memory.
  2: intros [H_abs H_interp]; contradiction H_abs; reflexivity.
  intros [H_triv H_smem1].
  elim H_state2_assign with System_Memory.
  2: intros [H_abs H_interp]; contradiction H_abs; reflexivity.
  intros [H_triv2 H_smem2].
  clear H_triv H_triv2.
  unfold state_component_interp in H_smem1, H_smem2.
  rewrite H_smem1. rewrite H_smem2.

  unfold drop_va_pt_page in H_drop1, H_drop2.
  destruct H_drop1 as [old_pt1 H_old_pt1_map [old_mp1 H_old_pt1_content H_newmem1]].
  destruct H_drop2 as [old_pt2 H_old_pt2_map [old_mp2 H_old_pt2_content H_newmem2]].

  intros pa' pg' [ma' H_ma'_hyp_map H_pg'_map] [val H_pg'_content].

  unfold current_vadd_mapping_madd, os_current_vadd_mapping_madd in H_pt_addr1_map, H_pt_addr2_map.
  destruct H_pt_addr1_map as [system H_system_map [p2m1 H_p2m1_map H_pt_addr1_map]].
  destruct H_pt_addr2_map as [system2 H_system2_map [p2m2 H_p2m2_map H_pt_addr2_map]].
  rewrite <- H_oss in H_system2_map.
  rewrite <- H_system_map in H_system2_map.
  injection H_system2_map; intro; subst system2; clear H_system2_map.

  (* ma' cannot be the changed pt_addr1, as it contains a RW page *)
  assert (ma'<>pt_addr1) as H_ma'_ne_pt_addr1.
  intro H_ma'_eq_pt_addr1.
  rewrite H_newmem1 in H_pg'_map.
  rewrite H_ma'_eq_pt_addr1 in H_pg'_map.
  rewrite lemma_map_add_apply in H_pg'_map.
  injection H_pg'_map; intro H_pg'_value;
    rewrite H_pg'_value in H_pg'_content; discriminate H_pg'_content.
  
  elim H_hyp1 with pa' pg'; auto.
  intros ma H_ma_map H_pg'_map2.
  exists ma; auto.

  unfold hypervisor_mapping_madd in H_ma_map.
  destruct H_ma_map as [p2m2' H_p2m2'_map H_ma_map].
  rewrite H_active_os in H_p2m2'_map.
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intro; subst p2m2'; clear H_p2m2'_map.

  assert (ma<>pt_addr2) as H_ma_ne_pt_addr2.
  intro H_ma_eq_pt_addr2.
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_injective_hyper H_va_has_pa.
  unfold valid_current_page in H_valid_currpg.

  elim H_valid_currpg with (st_active_os s2) system p2m2; auto.
  intros pt_addr2' H_pt_addr2'_map [H_pt_addr2'_type H_pt_addr2'_owner].

  rewrite <- H_pt_addr2_map in H_pt_addr2'_map.
  injection H_pt_addr2'_map; intro; subst pt_addr2'; clear H_pt_addr2'_map.
  rewrite <- H_ma_eq_pt_addr2 in H_pt_addr2'_type.
  unfold page_type in H_pt_addr2'_type.
  rewrite <- H_pg'_map2 in H_pt_addr2'_type.
  rewrite H_pg'_content in H_pt_addr2'_type.
  discriminate H_pt_addr2'_type.

  rewrite <- H_oss; assumption.
    
  rewrite H_newmem2.
  rewrite lemma_map_add_apply_distinct; auto.
  unfold padd_page.
  exists ma'; auto.
  rewrite H_newmem1 in H_pg'_map.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  exists val; auto.
Qed.

Lemma del_os_isolation_pt_pages : forall (s1 s2 s1' s2':state)
  (va:vadd) (pt_addr1 pt_addr2:madd),
  active_equivalent ctxt s1 s2 ->
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  current_vadd_mapping_madd s1 pt_addr1 -> current_vadd_mapping_madd s2 pt_addr2 ->
  post_all_del va pt_addr1 s1 s1' -> post_all_del va pt_addr2 s2 s2' ->
  osi_pt_pages_eq ctxt s1' s2' (st_active_os s1).
Proof.
  intros s1 s2 s1' s2' va pt_addr1 pt_addr2 H_active_eq H_s1_valid H_s2_valid H_pt_addr1_map H_pt_addr2_map.
  generalize H_active_eq.
  unfold active_equivalent;
    intros (H_active_os, (H_activity_aos,
      (H_oss, ((H_hyp1, H_hyp2), (H_page_s1s2, H_page_s2s1)))), H_post1, H_post2).
  generalize H_post1 H_post2.
  intros H_post1' H_post2'.
  unfold post_all_del in H_post1', H_post2'.
  destruct H_post1' as [newmem1 H_drop1 H_state1_assign].
  destruct H_post2' as [newmem2 H_drop2 H_state2_assign].

  unfold assign_system_memory, state_component_assign, state_component_update in H_state1_assign, H_state2_assign.

  unfold osi_pt_pages_eq, padd_page.

  (* Only System Memory changes *)
  assert (forall cmp, cmp <> System_Memory ->
    state_component_interp s1 cmp = state_component_interp s1' cmp)
    as H_eqcomp1.
  intros cmp H_eqcomp1; elim H_state1_assign with cmp.
  intros [H_abs H_mem].
  contradiction H_abs; auto.
  intros [H_eq1 H_interp1]; assumption.
  
  assert (forall cmp, cmp <> System_Memory ->
    state_component_interp s2 cmp = state_component_interp s2' cmp)
    as H_eqcomp2.
  intros cmp H_eqcomp2; elim H_state2_assign with cmp.
  intros [H_abs H_mem].
  contradiction H_abs; auto.
  intros [H_eq1 H_interp2]; assumption.
  
  Simp_Inv_ThreeComps H_eqcomp1 H_eqaos1 H_eqhyp1 H_eqoss1
    System_Memory Active_OS Hypervisor_SC Guest_OSs.
  rewrite <- H_eqhyp1.
  Simp_Inv_ThreeComps H_eqcomp2 H_eqaos2 H_eqhyp2 H_eqoss2
    System_Memory Active_OS Hypervisor_SC Guest_OSs.
  rewrite <- H_eqhyp2.
  elim H_state1_assign with System_Memory.
  2: intros [H_abs H_interp]; contradiction H_abs; reflexivity.
  intros [H_triv H_smem1].
  elim H_state2_assign with System_Memory.
  2: intros [H_abs H_interp]; contradiction H_abs; reflexivity.
  intros [H_triv2 H_smem2].
  clear H_triv H_triv2.
  unfold state_component_interp in H_smem1, H_smem2.
  rewrite H_smem1. rewrite H_smem2.

  unfold drop_va_pt_page in H_drop1, H_drop2.
  destruct H_drop1 as [old_pt1 H_old_pt1_map [old_mp1 H_old_pt1_content H_newmem1]].
  destruct H_drop2 as [old_pt2 H_old_pt2_map [old_mp2 H_old_pt2_content H_newmem2]].

  unfold current_vadd_mapping_madd, os_current_vadd_mapping_madd in H_pt_addr1_map, H_pt_addr2_map.
  destruct H_pt_addr1_map as [system H_system_map [p2m1 H_p2m1_map H_pt_addr1_map]].
  destruct H_pt_addr2_map as [system2 H_system2_map [p2m2 H_p2m2_map H_pt_addr2_map]].
  rewrite <- H_oss in H_system2_map.
  rewrite <- H_system_map in H_system2_map.
  injection H_system2_map; intro; subst system2; clear H_system2_map.
  unfold hypervisor_mapping in H_p2m1_map, H_p2m2_map.
  
  intros pt_padd pt_page1' mp1' [pt_addr1'
    H_pt_addr1'_map H_pt_page1'_map] H_pt_page1'_content.

  unfold hypervisor_mapping_madd in H_pt_addr1'_map.
  destruct H_pt_addr1'_map as [p2m1' H_p2m1'_map H_pt_addr1'_map].
  rewrite <- H_p2m1_map in H_p2m1'_map.
  injection H_p2m1'_map; intro; subst p2m1'; clear H_p2m1'_map.

  case (madd_eq pt_addr1 pt_addr1').
  intro H_pt_addr1'_eq_pt_addr1.

  assert (pt_padd=(os_current_page system)).
  apply hyp_mapped_same_paddr with ctxt s1 (st_active_os s1) pt_addr1; auto.
  rewrite H_pt_addr1'_eq_pt_addr1.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  subst pt_padd.

  rewrite H_newmem1 in H_pt_page1'_map.
  rewrite H_pt_addr1'_eq_pt_addr1 in H_pt_page1'_map.
  rewrite lemma_map_add_apply in H_pt_page1'_map.
  injection H_pt_page1'_map; intro H_pt_page1'_value.
  
  unfold hypervisor_mapping in H_p2m2_map.
  
  exists (Page (PT_c (map_drop vadd_eq old_mp2 va)) (page_owned_by old_pt2)).
  exists pt_addr2; auto.
  unfold hypervisor_mapping_madd.
  rewrite H_active_os.
  exists p2m2; auto.
  
  rewrite H_newmem2.
  rewrite lemma_map_add_apply.
  reflexivity.

  exists (map_drop vadd_eq old_mp2 va); auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va_accessible H_pg'_type H_ma'_hyp_map.

  rewrite H_pt_page1'_value in H_pt_page1'_content.
  simpl in H_pt_page1'_content.
  injection H_pt_page1'_content; intro H_mp1'_value.
  rewrite H_mp1'_value in H_ma'_map.

  assert (va<>va') as H_va_ne_va'.
  intro H_va_eq_va'.
  rewrite H_va_eq_va' in H_ma'_map.
  rewrite map_drop_apply in H_ma'_map.
  discriminate H_ma'_map.

  rewrite map_drop_apply_distinct in H_ma'_map; auto.
  rewrite map_drop_apply_distinct; auto.

  unfold hypervisor_mapping_madd in H_ma'_hyp_map.
  destruct H_ma'_hyp_map as [p2m1' H_p2m1'_map H_ma'_hyp_map].
  rewrite <- H_p2m1_map in H_p2m1'_map.
  injection H_p2m1'_map; intro; subst p2m1'; clear H_p2m1'_map.

  assert ((os_current_page system) <> pa') as H_pg'_not_curr_pt.
  intro H_pa'_curr_pt.
  assert (pt_addr1=ma') as H_pt_addr1_eq_ma'.
  apply hyp_mapped_same_maddr with s1 (st_active_os s1) (os_current_page system).
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  rewrite H_pa'_curr_pt.
  assumption.
  rewrite <- H_pt_addr1_eq_ma' in H_pg'_type.
  unfold page_type in H_pg'_type.
  rewrite H_newmem1 in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_type.
  simpl in H_pg'_type.
  discriminate H_pg'_type.

  assert (pt_addr1<>ma') as H_pt_addr1_ne_ma1.
  intro H_pt_addr1_eq_ma'.
  apply H_pg'_not_curr_pt.
  apply hyp_mapped_same_paddr with ctxt s1 (st_active_os s1) pt_addr1; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  rewrite H_pt_addr1_eq_ma'.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.


  elim H_page_s1s2 with (os_current_page system) old_pt1 old_mp1; auto.
  intros pt_page2 H_pt_page2_map [mp2 H_pt_page2_content H_virt_map_eq].
  unfold padd_page in H_pt_page2_map.
  destruct H_pt_page2_map as [pt_addr2' H_pt_addr2'_map H_pt_page2_map].
  unfold hypervisor_mapping_madd in H_pt_addr2'_map.
  destruct H_pt_addr2'_map as [p2m2' H_p2m2'_map H_pt_addr2'_map].
  rewrite H_active_os in H_p2m2'_map.
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intro; subst p2m2'; clear H_p2m2'_map.
  rewrite <- H_pt_addr2_map in H_pt_addr2'_map.
  injection H_pt_addr2'_map; intro; subst pt_addr2'; clear H_pt_addr2'_map.

  elim H_virt_map_eq with va' pa' ma' pg'; auto.
  intros ma2 H_ma2_map [H_ma2_hyp_map H_pg'_map2].
  exists ma2; auto.

  rewrite <- H_old_pt2_map in H_pt_page2_map.
  injection H_pt_page2_map; intro; subst pt_page2; clear H_pt_page2_map.
  rewrite <- H_old_pt2_content in H_pt_page2_content.
  injection H_pt_page2_content; intro; subst mp2.
  assumption.

  split; auto.
  rewrite H_newmem2.

  assert (pt_addr2<>ma2) as H_pt_addr2_ne_ma2.
  intro H_pt_addr2_eq_ma2.
  apply H_pg'_not_curr_pt.
  apply hyp_mapped_same_paddr with ctxt s2 (st_active_os s2) pt_addr2; auto.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.
  rewrite <- H_active_os; auto.
  rewrite H_pt_addr2_eq_ma2.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.
  rewrite H_active_os; assumption.

  unfold hypervisor_mapping_madd in H_ma2_hyp_map.
  destruct H_ma2_hyp_map as [p2m2' H_p2m2'_map H_ma2_hyp_map].
  rewrite H_active_os in H_p2m2'_map.
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intro; subst p2m2'; clear H_p2m2'_map.
  assumption.

  rewrite lemma_map_add_apply_distinct; auto.

  rewrite H_newmem1 in H_pg'_map.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  rewrite H_newmem1 in H_pg'_type.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_type; auto.

  unfold hypervisor_mapping_madd; exists p2m1; auto.

  unfold padd_page.
  exists pt_addr1; auto.
  unfold hypervisor_mapping_madd; exists p2m1; auto.


  (* If pt_addr1' is not the current PT page *)
  intro H_pt_addr1'_ne_pt_addr1.

  elim H_page_s1s2 with pt_padd pt_page1' mp1'; auto.
  intros pt_page2 H_pt_page2_map [mp2 H_pt_page2_content H_virt_map_eq].
  unfold padd_page in H_pt_page2_map.
  destruct H_pt_page2_map as [pt_addr2' H_pt_addr2'_map H_pt_page2_map].
  unfold hypervisor_mapping_madd in H_pt_addr2'_map.
  destruct H_pt_addr2'_map as [p2m2' H_p2m2'_map H_pt_addr2'_map].
  rewrite H_active_os in H_p2m2'_map.
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intro; subst p2m2'; clear H_p2m2'_map.
  
  assert (pt_padd<>(os_current_page system)) as H_pt_padd_ne_curr_pt.
  intro H_pt_padd_eq_curr_pt.
  apply H_pt_addr1'_ne_pt_addr1.
  apply hyp_mapped_same_maddr with s1 (st_active_os s1) pt_padd.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  rewrite H_pt_padd_eq_curr_pt.
  assumption.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  
  assert (pt_addr2'<>pt_addr2) as H_pt_addr2'_ne_pt_addr2.
  intro H_pt_addr2'_eq_pt_addr2.
  apply H_pt_padd_ne_curr_pt.
  apply hyp_mapped_same_paddr with ctxt s2 (st_active_os s2) pt_addr2; auto.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.
  rewrite <- H_pt_addr2'_eq_pt_addr2.
  assumption.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.

  exists pt_page2.
  exists pt_addr2'.
  unfold hypervisor_mapping_madd.
  rewrite H_active_os.
  exists p2m2; auto.

  rewrite H_newmem2.
  rewrite lemma_map_add_apply_distinct; auto.
  exists mp2; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map
    H_va_accessible H_pg'_type H_ma'_hyp_map.

  rewrite H_newmem1 in H_pg'_map, H_pg'_type.
  unfold page_type in H_pg'_type.

  assert (pt_addr1<>ma') as H_pt_addr1_neq_ma'.
  intro H_pt_addr1_eq_ma'.
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_injective_hyper H_va_has_pa.
  unfold valid_current_page in H_valid_currpg.
  elim H_valid_currpg with (st_active_os s1) system p2m1; auto.
  intros pt_addr1'' H_pt_addr1''_map [H_pt_addr1_type H_pt_addr1_owner].
  rewrite <- H_pt_addr1_map in H_pt_addr1''_map.
  injection H_pt_addr1''_map; intro; subst pt_addr1''; clear H_pt_addr1''_map.
  rewrite H_pt_addr1_eq_ma' in H_pt_addr1_type.
  rewrite H_pt_addr1_eq_ma' in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_type.
  simpl in H_pg'_type.
  discriminate H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_map, H_pg'_type; auto.
  rewrite <- H_pg'_map in H_pg'_type.

  elim H_virt_map_eq with va' pa' ma' pg'; auto.
  intros ma H_ma_map [H_ma_hyp_map H_pg'_map2].
  exists ma; auto.
  split; auto.
  
  rewrite H_newmem2.
  assert (pt_addr2<>ma) as H_pt_addr2_neq_ma.
  intro H_pt_addr2_eq_ma.
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_injective_hyper H_va_has_pa.
  unfold valid_current_page in H_valid_currpg.
  elim H_valid_currpg with (st_active_os s2) system p2m2; auto.
  intros pt_addr2'' H_pt_addr2''_map [H_pt_addr2_type H_pt_addr2_owner].
  rewrite <- H_pt_addr2_map in H_pt_addr2''_map.
  injection H_pt_addr2''_map; intro; subst pt_addr2''; clear H_pt_addr2''_map.
  rewrite H_pt_addr2_eq_ma in H_pt_addr2_type.
  unfold page_type in H_pt_addr2_type.
  rewrite <- H_pg'_map2 in H_pt_addr2_type.
  rewrite H_pg'_type in H_pt_addr2_type.
  discriminate H_pt_addr2_type.

  rewrite <- H_oss; auto.

  rewrite lemma_map_add_apply_distinct; auto.
  unfold page_type.
  rewrite <- H_pg'_map.
  rewrite H_pg'_type; reflexivity.
  unfold padd_page.
  exists pt_addr1'.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  rewrite H_newmem1 in H_pt_page1'_map.
  rewrite lemma_map_add_apply_distinct in H_pt_page1'_map; auto.
Qed.

Lemma del_os_isolation : forall (s1 s2 s1' s2':state)
  (va:vadd) (pt_addr1 pt_addr2:madd),
  active_equivalent ctxt s1 s2 ->
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  current_vadd_mapping_madd s1 pt_addr1 -> current_vadd_mapping_madd s2 pt_addr2 ->
  post_all_del va pt_addr1 s1 s1' -> post_all_del va pt_addr2 s2 s2' ->
  active_equivalent ctxt s1' s2'.
Proof.
  intros s1 s2 s1' s2' va pt_addr1 pt_addr2 H_active_eq H_s1_valid H_s2_valid H_pt_addr1_map H_pt_addr2_map.
  generalize H_active_eq.
  unfold active_equivalent;
    intros (H_active_os, (H_activity_aos,
      (H_oss, ((H_hyp1, H_hyp2), (H_page_s1s2, H_page_s2s1)))), H_post1, H_post2).
  generalize H_post1 H_post2.
  intros H_post1' H_post2'.
  unfold post_all_new in H_post1', H_post2'.
  destruct H_post1' as [newmem1 H_drop1 H_state1_assign].
  destruct H_post2' as [newmem2 H_drop2 H_state2_assign].

  unfold assign_system_memory, state_component_assign, state_component_update in H_state1_assign, H_state2_assign.

  unfold aos_equivalent_hyper_mapping, aos_equivalent_pt_pages,
    osi_equivalent_hyper_mapping, osi_equivalent_pt_pages.

  (* Only System Memory changes *)
  assert (forall cmp, cmp <> System_Memory ->
    state_component_interp s1 cmp = state_component_interp s1' cmp)
    as H_eqcomp1.
  intros cmp H_eqcomp1; elim H_state1_assign with cmp.
  intros [H_abs H_mem].
  contradiction H_abs; auto.
  intros [H_eq1 H_interp1]; assumption.

  assert (forall cmp, cmp <> System_Memory ->
    state_component_interp s2 cmp = state_component_interp s2' cmp)
    as H_eqcomp2.
  intros cmp H_eqcomp2; elim H_state2_assign with cmp.
  intros [H_abs H_mem].
  contradiction H_abs; auto.
  intros [H_eq1 H_interp2]; assumption.
  Simp_Inv_ThreeComps H_eqcomp1 H_eqaos1 H_eqact1 H_eqoss1
    System_Memory Active_OS OS_Activity Guest_OSs.
  rewrite <- H_eqaos1.
  rewrite <- H_eqact1.
  rewrite <- H_eqoss1.
  Simp_Inv_ThreeComps H_eqcomp2 H_eqaos2 H_eqact2 H_eqoss2
    System_Memory Active_OS OS_Activity Guest_OSs.
  rewrite <- H_eqaos2.
  rewrite <- H_eqact2.
  rewrite <- H_eqoss2.

  generalize (active_eq_symm ctxt s1 s2); intro H_s2_equiv_s1.
  
  repeat split; try assumption.
 
  apply del_os_isolation_hyper with s2 va pt_addr1 pt_addr2; auto.
  rewrite H_active_os.
  apply del_os_isolation_hyper with s1 va pt_addr2 pt_addr1; auto.
  apply del_os_isolation_pt_pages with s2 va pt_addr1 pt_addr2; auto.
  rewrite H_active_os.
  apply del_os_isolation_pt_pages with s1 va pt_addr2 pt_addr1; auto.
Qed.


Lemma lswitch_os_isolation : forall (s1 s2 s1' s2':state) (osi:os_ident) (pa:padd),
  active_equivalent ctxt s1 s2 ->
  post_all_lswitch osi pa None s1 s1' -> post_all_lswitch osi pa None s2 s2' ->
  active_equivalent ctxt s1' s2'.
Proof.
  intros s1 s2 s1' s2' osi pa.
  unfold active_equivalent, aos_equivalent_hyper_mapping,
    aos_equivalent_pt_pages, padd_page, vadd_page;
    intros (H_active_os, (H_activity_aos,
      (H_oss, ((H_hyp1, H_hyp2), (H_page_s1s2, H_page_s2s1)))), H_post1, H_post2).

  unfold post_all_lswitch in H_post1, H_post2.
  destruct H_post1 as [os1 H_os1_map H_state1_assign].
  destruct H_post2 as [os2 H_os2_map H_state2_assign].
  unfold state_component_assign, state_component_update
    in H_state1_assign, H_state2_assign.

  unfold aos_equivalent_hyper_mapping, aos_equivalent_pt_pages,
    osi_equivalent_hyper_mapping, osi_equivalent_pt_pages,
    osi_hyper_mapping_eq, osi_pt_pages_eq, padd_page in *.

  (* Only Guest_OSs component changes *)
  assert (forall cmp, cmp <> Guest_OSs ->
    state_component_interp s1 cmp = state_component_interp s1' cmp)
    as H_eqcomp1.
  intros cmp H_eqcomp; elim H_state1_assign with cmp.
  induction cmp; try (intros [H_abs H]; discriminate);
    try (contradiction H_eqcomp; reflexivity).
  intros [H_eqcomp' H_interp1]; assumption.
  assert (forall cmp, cmp <> Guest_OSs ->
    state_component_interp s2 cmp = state_component_interp s2' cmp)
    as H_eqcomp2.
  intros cmp H_eqcomp; elim H_state2_assign with cmp.
  induction cmp; try (intros [H_abs H]; discriminate);
    try (contradiction H_eqcomp; reflexivity).
  intros [H_eqcomp' H_interp2]; assumption.
  Simp_Inv_ThreeComps H_eqcomp1 H_eqaos1 H_eqact1 H_eqhyp1
    Guest_OSs Active_OS OS_Activity Hypervisor_SC.
  rewrite <- H_eqaos1.
  rewrite <- H_eqact1.
  rewrite <- H_eqhyp1.
  Simp_Inv_OneComp H_eqcomp1 H_eqsmem1 Guest_OSs System_Memory.
  rewrite <- H_eqsmem1.
  Simp_Inv_ThreeComps H_eqcomp2 H_eqaos2 H_eqact2 H_eqhyp2
    Guest_OSs Active_OS OS_Activity Hypervisor_SC.
  rewrite <- H_eqaos2.
  rewrite <- H_eqact2.
  rewrite <- H_eqhyp2.
  Simp_Inv_OneComp H_eqcomp2 H_eqsmem2 Guest_OSs System_Memory.
  rewrite <- H_eqsmem2.
  
  repeat split; try assumption.
  
  unfold state_component_interp in H_state1_assign, H_state2_assign.
  elim H_state1_assign with Guest_OSs;
    try (intros [H_abs H']; contradiction H_abs; reflexivity).
  intros [H_cmp1 H_oss1']; rewrite H_oss1'.
  elim H_state2_assign with Guest_OSs;
    try (intros [H_abs H']; contradiction H_abs; reflexivity).
  intros [H_cmp2 H_oss2']; rewrite H_oss2'.
  rewrite <- H_active_os.
  case (os_ident_eq osi (st_active_os s1)).
  intro H_osi.
  rewrite H_osi.
  repeat rewrite lemma_map_add_apply; reflexivity.
  intros H_o_neq_aos.
  repeat rewrite lemma_map_add_apply_distinct; auto.
  rewrite <- H_active_os in H_oss.
  assumption.
Qed.

Lemma hcall_os_isolation : forall (s1 s2 s1' s2':state) (hc:Hyperv_call),
  active_equivalent ctxt s1 s2 ->
  post_hcall hc s1 s1' -> post_hcall hc s2 s2' ->
  active_equivalent ctxt s1' s2'.
Proof.
  intros s1 s2 s1' s2' hc.
  unfold active_equivalent, aos_equivalent_hyper_mapping,
    aos_equivalent_pt_pages, padd_page, vadd_page;
    intros (H_active_os, (H_activity_aos,
      (H_oss, ((H_hyp1, H_hyp2), (H_page_s1s2, H_page_s2s1)))), H_post1, H_post2).

  unfold post_hcall in H_post1, H_post2.
  destruct H_post1 as [aos1 H_aos1_map H_s1'].
  destruct H_post2 as [aos2 H_aos2_map H_s2'].

  rewrite H_s1'; rewrite H_s2'; simpl.

  repeat split; try reflexivity; try assumption.

  repeat rewrite lemma_map_add_apply. 
  rewrite <- H_oss in H_aos2_map.
  rewrite <- H_aos1_map in H_aos2_map.
  injection H_aos2_map.
  intros H_aos2_eq_aos1; rewrite H_aos2_eq_aos1.
  reflexivity.
Qed.

Lemma ret_ctrl_os_isolation : forall (s1 s2 s1' s2':state),
  active_equivalent ctxt s1 s2 ->
  post_ret_ctrl s1 s1' -> post_ret_ctrl s2 s2' ->
  active_equivalent ctxt s1' s2'.
Proof.
  intros s1 s2 s1' s2'.
  unfold active_equivalent,
    aos_equivalent_hyper_mapping, aos_equivalent_pt_pages,
    osi_equivalent_hyper_mapping, osi_equivalent_pt_pages,
    osi_hyper_mapping_eq, osi_pt_pages_eq, padd_page, vadd_page;
    intros (H_active_os, (H_activity_aos,
      (H_oss, ((H_hyp1, H_hyp2), (H_page_s1s2, H_page_s2s1)))), H_post1, H_post2).
  unfold post_ret_ctrl in H_post1, H_post2.
  unfold state_components_assign, state_components_update
    in H_post1, H_post2.

  (* Only Exec_Mode and OS_Activity components change *)
  assert (forall cmp, cmp <> Exec_Mode -> cmp <> OS_Activity ->
    state_component_interp s1 cmp = state_component_interp s1' cmp)
    as H_eqcomp1.
  intros cmp H_eqcomp1 H_eqcomp1'; elim H_post1 with cmp.
  intros [[H_eq1 H_eq2] H_interp1]; assumption.
  intros [H_abs [H_exec H_activity]].
  contradiction H_abs; auto.
  assert (forall cmp, cmp <> Exec_Mode -> cmp <> OS_Activity ->
    state_component_interp s2 cmp = state_component_interp s2' cmp)
    as H_eqcomp2.
  intros cmp H_eqcomp2 H_eqcomp2'; elim H_post2 with cmp.
  intros [[H_eq1 H_eq2] H_interp2]; assumption.
  intros [H_abs [H_exec H_activity]].
  contradiction H_abs; auto.
  Simp_Inv_ThreeComps H_eqcomp1 H_eqaos1 H_eqhyp1 H_eqoss1
    Exec_Mode Active_OS Hypervisor_SC Guest_OSs.
  rewrite <- H_eqaos1; try (intros; discriminate).
  rewrite <- H_eqhyp1; try (intros; discriminate).
  rewrite <- H_eqoss1;  try (intros; discriminate).
  Simp_Inv_OneComp H_eqcomp1 H_eqsmem1 Exec_Mode System_Memory.
  rewrite <- H_eqsmem1; try (intros; discriminate).
  Simp_Inv_ThreeComps H_eqcomp2 H_eqaos2 H_eqhyp2 H_eqoss2
    Exec_Mode Active_OS Hypervisor_SC Guest_OSs.
  rewrite <- H_eqaos2; try (intros; discriminate).
  rewrite <- H_eqhyp2; try (intros; discriminate).
  rewrite <- H_eqoss2;  try (intros; discriminate).
  Simp_Inv_OneComp H_eqcomp2 H_eqsmem2 Exec_Mode System_Memory.
  rewrite <- H_eqsmem2; try (intros; discriminate).

  repeat split; try assumption.
  
  elim H_post1 with OS_Activity.
  intros [[H_eqcmp H_abs] H_interp]; contradiction H_abs; reflexivity.
  intros [H_eqcmp1 [H_exec1 H_activity1]].
  elim H_post2 with OS_Activity.
  intros [[H_eqcmp2 H_abs] H_interp]; contradiction H_abs; reflexivity.
  intros [H_eqcmp2 [H_exec2 H_activity2]].  
  unfold state_component_interp in H_activity1, H_activity2.
  rewrite H_activity1; rewrite H_activity2; reflexivity.
Qed.

Lemma chmod_os_isolation : forall (s1 s2 s1' s2':state),
  active_equivalent ctxt s1 s2 ->
  post_chmod ctxt s1 s1' -> post_chmod ctxt s2 s2' ->
  active_equivalent ctxt s1' s2'.
Proof.
  intros s1 s2 s1' s2'.
  unfold active_equivalent,
    aos_equivalent_hyper_mapping, aos_equivalent_pt_pages,
    osi_equivalent_hyper_mapping, osi_equivalent_pt_pages,
    osi_hyper_mapping_eq, osi_pt_pages_eq, padd_page, vadd_page;
    intros (H_active_os, (H_activity_aos,
      (H_oss, ((H_hyp1, H_hyp2), (H_page_s1s2, H_page_s2s1)))), H_post1, H_post2).
  unfold post_chmod in H_post1, H_post2.
  destruct H_post1 as [[H_trusted_aos1 H_state1_assign] | [H_untrusted_aos1 H_state1_assign]];
    destruct H_post2 as [[H_trusted_aos2 H_state2_assign] | [H_untrusted_aos2 H_state2_assign]].

  (* trusted aos *)
  unfold state_components_assign, state_components_update
    in H_state1_assign, H_state2_assign.

  (* Only Exec_Mode and OS_Activity components change *)
  assert (forall cmp, cmp <> Exec_Mode -> cmp <> OS_Activity ->
    state_component_interp s1 cmp = state_component_interp s1' cmp)
    as H_eqcomp1.
  intros cmp H_eqcomp1 H_eqcomp1'; elim H_state1_assign with cmp.
  intros [[H_eq1 H_eq2] H_interp1]; assumption.
  intros [H_abs [H_exec H_activity]].
  contradiction H_abs; auto.
  assert (forall cmp, cmp <> Exec_Mode -> cmp <> OS_Activity ->
    state_component_interp s2 cmp = state_component_interp s2' cmp)
    as H_eqcomp2.
  intros cmp H_eqcomp2 H_eqcomp2'; elim H_state2_assign with cmp.
  intros [[H_eq1 H_eq2] H_interp2]; assumption.
  intros [H_abs [H_exec H_activity]].
  contradiction H_abs; auto.
  Simp_Inv_ThreeComps H_eqcomp1 H_eqaos1 H_eqhyp1 H_eqoss1
    Exec_Mode Active_OS Hypervisor_SC Guest_OSs.
  rewrite <- H_eqaos1; try (intros; discriminate).
  rewrite <- H_eqhyp1; try (intros; discriminate).
  rewrite <- H_eqoss1;  try (intros; discriminate).
  Simp_Inv_OneComp H_eqcomp1 H_eqsmem1 Exec_Mode System_Memory.
  rewrite <- H_eqsmem1; try (intros; discriminate).
  Simp_Inv_ThreeComps H_eqcomp2 H_eqaos2 H_eqhyp2 H_eqoss2
    Exec_Mode Active_OS Hypervisor_SC Guest_OSs.
  rewrite <- H_eqaos2; try (intros; discriminate).
  rewrite <- H_eqhyp2; try (intros; discriminate).
  rewrite <- H_eqoss2;  try (intros; discriminate).
  Simp_Inv_OneComp H_eqcomp2 H_eqsmem2 Exec_Mode System_Memory.
  rewrite <- H_eqsmem2; try (intros; discriminate).

  repeat split; try assumption.

  elim H_state1_assign with OS_Activity.
  intros [[H_eqcmp H_abs] H_interp]; contradiction H_abs; reflexivity.
  intros [H_eqcmp1 [H_exec1 H_activity1]].
  elim H_state2_assign with OS_Activity.
  intros [[H_eqcmp2 H_abs] H_interp]; contradiction H_abs; reflexivity.
  intros [H_eqcmp2 [H_exec2 H_activity2]].  
  unfold state_component_interp in H_activity1, H_activity2.
  rewrite H_activity1; rewrite H_activity2; reflexivity.
  
  unfold trusted_aos in H_trusted_aos1, H_untrusted_aos2.
  rewrite <- H_active_os in H_untrusted_aos2.
  contradiction H_untrusted_aos2.
  unfold trusted_aos in H_untrusted_aos1, H_trusted_aos2.
  rewrite H_active_os in H_untrusted_aos1.
  contradiction H_untrusted_aos1.

  (* Untrusted OS *)
  unfold state_components_assign, state_components_update
    in H_state1_assign, H_state2_assign.

  (* Only Exec_Mode and OS_Activity components change *)
  assert (forall cmp, cmp <> Exec_Mode -> cmp <> OS_Activity ->
    state_component_interp s1 cmp = state_component_interp s1' cmp)
    as H_eqcomp1.
  intros cmp H_eqcomp1 H_eqcomp1'; elim H_state1_assign with cmp.
  intros [[H_eq1 H_eq2] H_interp1]; assumption.
  intros [H_abs [H_exec H_activity]].
  contradiction H_abs; auto.
  assert (forall cmp, cmp <> Exec_Mode -> cmp <> OS_Activity ->
    state_component_interp s2 cmp = state_component_interp s2' cmp)
    as H_eqcomp2.
  intros cmp H_eqcomp2 H_eqcomp2'; elim H_state2_assign with cmp.
  intros [[H_eq1 H_eq2] H_interp2]; assumption.
  intros [H_abs [H_exec H_activity]].
  contradiction H_abs; auto.
  Simp_Inv_ThreeComps H_eqcomp1 H_eqaos1 H_eqhyp1 H_eqoss1
    Exec_Mode Active_OS Hypervisor_SC Guest_OSs.
  rewrite <- H_eqaos1; try (intros; discriminate).
  rewrite <- H_eqhyp1; try (intros; discriminate).
  rewrite <- H_eqoss1;  try (intros; discriminate).
  Simp_Inv_OneComp H_eqcomp1 H_eqsmem1 Exec_Mode System_Memory.
  rewrite <- H_eqsmem1; try (intros; discriminate).
  Simp_Inv_ThreeComps H_eqcomp2 H_eqaos2 H_eqhyp2 H_eqoss2
    Exec_Mode Active_OS Hypervisor_SC Guest_OSs.
  rewrite <- H_eqaos2; try (intros; discriminate).
  rewrite <- H_eqhyp2; try (intros; discriminate).
  rewrite <- H_eqoss2;  try (intros; discriminate).
  Simp_Inv_OneComp H_eqcomp2 H_eqsmem2 Exec_Mode System_Memory.
  rewrite <- H_eqsmem2; try (intros; discriminate).

  repeat split; try assumption.

  elim H_state1_assign with OS_Activity.
  intros [[H_eqcmp H_abs] H_interp]; contradiction H_abs; reflexivity.
  intros [H_eqcmp1 [H_exec1 H_activity1]].
  elim H_state2_assign with OS_Activity.
  intros [[H_eqcmp2 H_abs] H_interp]; contradiction H_abs; reflexivity.
  intros [H_eqcmp2 [H_exec2 H_activity2]].  
  unfold state_component_interp in H_activity1, H_activity2.
  rewrite H_activity1; rewrite H_activity2; reflexivity.
Qed.

Lemma pin_trusted_os_isolation_hyper : forall (s1 s2 s1' s2':state) (pa:padd) (type:ptype),
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  active_equivalent ctxt s1 s2 ->
  valid_exec ctxt s1 (Page_pin_trusted pa type) s1' ->
  valid_exec ctxt s2 (Page_pin_trusted pa type) s2' ->
  osi_hyper_mapping_eq s1' s2' (st_active_os s1').
Proof.
  intros s1 s2 s1' s2' pa type H_s1_valid H_s2_valid.
  unfold active_equivalent, aos_equivalent_hyper_mapping,
    aos_equivalent_pt_pages, osi_equivalent_hyper_mapping, osi_equivalent_pt_pages;
    intros (H_active_os, (H_activity_aos,
      (H_oss, ((H_hyp1, H_hyp2), (H_page_s1s2, H_page_s2s1)))), H_exec1, H_exec2).
  generalize H_exec1; generalize H_exec2.
  intros H_exec2' H_exec1'.
  inversion_clear H_exec1'.
  inversion_clear H_exec2'.
  rename H into H_pre1.
  rename H0 into H_post1.
  rename H1 into H_pre2.
  rename H2 into H_post2.
  unfold post_action, post_page_pin_trusted in H_post1, H_post2.
  unfold post_all_pin in H_post1, H_post2.
  destruct H_post1 as [ma1 [pg1 H_pg1_map [H_pg1_owner
    [H_pg1_content [p2m1 H_post_hyp1 [os1 H_os1_map H_s1']]]]]].
  destruct H_post2 as [ma2 [pg2 H_pg2_map [H_pg2_owner
    [H_pg2_content [p2m2 H_post_hyp2 [os2 H_os2_map H_s2']]]]]].

  unfold osi_hyper_mapping_eq, padd_page.
  rewrite H_s1'; rewrite H_s2'; simpl.
  intros pa' pg' H_pg'_map [val H_pg'_content].
  case (padd_eq pa pa').
  (* If pa' is the new pa, it changes to the same page *)
  intros; subst pa'.
  unfold hypervisor_mapping_madd.
  exists ma2.
  exists (map_add padd_eq p2m2 pa ma2).
  rewrite H_active_os.
  rewrite lemma_map_add_apply; auto.
  rewrite lemma_map_add_apply; auto.  

  rewrite lemma_map_add_apply.
  destruct H_pg'_map as [ma H_hyp_pa_ma1 H_pg'_value].
  assert (ma = ma1).
  unfold hypervisor_mapping_madd in H_hyp_pa_ma1.
  destruct H_hyp_pa_ma1 as [p2m1' H_p2m1'_map H_ma_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intros H_p2m1'_map'.
  rewrite H_p2m1'_map' in H_ma_map.
  rewrite lemma_map_add_apply in H_ma_map; injection H_ma_map; auto.
  subst ma.

  induction (type).
  (* The pinned page is a PT. This case is absurd. *)
  rewrite lemma_map_add_apply in H_pg'_value.
  injection H_pg'_value; intro H_pg'.
  rewrite H_pg' in H_pg'_content.
  simpl in H_pg'_content.
  discriminate H_pg'_content.
  (* The pinned page has RW type *)
  rewrite lemma_map_add_apply in H_pg'_value.
  rewrite <- H_active_os; assumption.

  (* If pa' is not the new pa, it does not change. *)
  intro H_pa_ne_pa'.

  assert (valid_state ctxt s1') as H_s1'_valid.
  apply valid_state_invariant with s1 (Page_pin_trusted pa type); auto.
  generalize H_s1'_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_valid_currpg H_va_has_pa.
  rename H_injective_hyper into H_inj_hyp_s1'.

  unfold hypervisor_mapping_madd.
  rewrite <- H_active_os.
  rewrite lemma_map_add_apply.
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  unfold hypervisor_mapping in H_post_hyp2.

  destruct H_pg'_map as [ma' H_hyp_pa_ma1 H_pg'_value].
  assert (ma1 <> ma').
  intros H_ma1_eq_ma'.
  apply H_pa_ne_pa'.

  unfold injective_hyper_mappings in H_inj_hyp_s1'.
  rewrite H_inj_hyp_s1' with (st_active_os s1) (map_add padd_eq p2m1 pa ma1) pa pa'; auto.

  rewrite H_s1'; simpl.
  rewrite lemma_map_add_apply.
  reflexivity.

  unfold hypervisor_mapping_madd in H_hyp_pa_ma1.
  destruct H_hyp_pa_ma1 as [p2m1' H_p2m1'_map H_ma'_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; clear H_p2m1'_map; intros H_p2m1'_map.
  rewrite lemma_map_add_apply.
  rewrite <- H_p2m1'_map.
  rewrite H_ma1_eq_ma'.
  assumption.

  rewrite lemma_map_add_apply_distinct in H_pg'_value; auto.

  destruct (H_valid_hyper (st_active_os s2) pa' p2m2 H_post_hyp2)
    as [[ma H_ma_map H_ma_owner]|H_error].
  exists ma.
  exists (map_add padd_eq p2m2 pa ma2); auto.
  rewrite lemma_map_add_apply_distinct; auto.
  assert (ma <> ma2).
  intros H_ma_eq_ma2.
  apply H_pa_ne_pa'.

  assert (valid_state ctxt s2') as H_s2'_valid.
  apply valid_state_invariant with s2 (Page_pin_trusted pa type); auto.
  rename H_valid_hyper into H_valid_hyper_s2.
  generalize H_s2'_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_valid_currpg H_va_has_pa.
  unfold injective_hyper_mappings in H_injective_hyper.
  rewrite H_injective_hyper with (st_active_os s2) (map_add padd_eq p2m2 pa ma2) pa pa'; auto.
  
  rewrite H_s2'; simpl.
  rewrite lemma_map_add_apply.
  reflexivity.
  cut (Value padd ma = map_apply padd_eq (map_add padd_eq p2m2 pa ma2) pa').
  intros H_ma_map'.
  rewrite <- H_ma_map'.
  rewrite lemma_map_add_apply.
  subst ma; reflexivity.
  rewrite lemma_map_add_apply_distinct; auto.

  rewrite lemma_map_add_apply_distinct; auto.

  elim H_hyp1 with pa' pg'.
  intros max H_max_map.
  unfold hypervisor_mapping_madd in H_max_map.
  assert (max=ma).
  destruct H_max_map as [p2m2' H_p2m2'_map H_max_map].
  rewrite H_active_os in H_p2m2'_map.
  rewrite <- H_p2m2'_map in H_post_hyp2.
  injection H_post_hyp2.
  intros H_p2m2_eq_p2m2'; subst p2m2'.
  rewrite <- H_max_map in H_ma_map.
  injection H_ma_map; auto.
  subst max.
  auto.

  unfold padd_page.
  exists ma'; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  unfold hypervisor_mapping_madd in H_hyp_pa_ma1.
  destruct H_hyp_pa_ma1 as [p2m1' H_p2m1'_map H_ma'_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; clear H_p2m1'_map;intros H_p2m1'_map.
  rewrite H_p2m1'_map in H_ma'_map.
  rewrite lemma_map_add_apply_distinct in H_ma'_map; auto.
  exists val; auto.
  
  elim H_hyp1 with pa' pg'.
  intros ma H_ma_map H_pg'_map2.
  unfold hypervisor_mapping_madd in H_ma_map.
  destruct H_ma_map as [p2m2' H_p2m2'_map H_ma_map].
  rewrite H_active_os in H_p2m2'_map.
  rewrite <- H_p2m2'_map in H_post_hyp2.
  injection H_post_hyp2.
  intros H_p2m2_eq_p2m2'; subst p2m2'.
  rewrite H_error in H_ma_map.
  discriminate H_ma_map.

  unfold padd_page, hypervisor_mapping_madd.
  exists ma'; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  unfold hypervisor_mapping_madd in H_hyp_pa_ma1.
  destruct H_hyp_pa_ma1 as [p2m1' H_p2m1'_map H_ma'_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; clear H_p2m1'_map;intros H_p2m1'_map.
  rewrite H_p2m1'_map in H_ma'_map.
  rewrite lemma_map_add_apply_distinct in H_ma'_map; auto.
  exists val; auto.
Qed.  

Lemma pin_trusted_os_isolation_pt_pages : forall (s1 s2 s1' s2':state) (pa:padd) (type:ptype),
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  active_equivalent ctxt s1 s2 ->
  valid_exec ctxt s1 (Page_pin_trusted pa type) s1' ->
  valid_exec ctxt s2 (Page_pin_trusted pa type) s2' ->
  osi_pt_pages_eq ctxt s1' s2' (st_active_os s1').
Proof.
  intros s1 s2 s1' s2' pa type H_s1_valid H_s2_valid.
  unfold active_equivalent, aos_equivalent_hyper_mapping,
    aos_equivalent_pt_pages, osi_equivalent_pt_pages, osi_pt_pages_eq;
    intros (H_active_os, (H_activity_aos,
      (H_oss, ((H_hyp1, H_hyp2), (H_page_s1s2, H_page_s2s1)))), H_exec1, H_exec2).
  generalize H_exec1; generalize H_exec2.
  intros H_exec2' H_exec1'.
  inversion_clear H_exec1'.
  inversion_clear H_exec2'.
  rename H into H_pre1.
  rename H0 into H_post1.
  rename H1 into H_pre2.
  rename H2 into H_post2.
  unfold post_action, post_page_pin_trusted in H_post1, H_post2.
  unfold post_all_pin in H_post1, H_post2.
  destruct H_post1 as [ma1 [pg1 H_pg1_map [H_pg1_owner
    [H_pg1_content [p2m1 H_post_hyp1 [os1 H_os1_map H_s1']]]]]].
  destruct H_post2 as [ma2 [pg2 H_pg2_map [H_pg2_owner
    [H_pg2_content [p2m2 H_post_hyp2 [os2 H_os2_map H_s2']]]]]].
  unfold hypervisor_mapping in H_post_hyp1, H_post_hyp2.

  unfold pre_action, pre_page_pin_trusted, pre_all_pin in H_pre1, H_pre2.
  destruct H_pre1 as [H_os_running1 [H_trusted_aos1 [[p2m1' H_p2m1'_map H_pa_err1] H_pre1]]].
  rewrite <- H_post_hyp1 in H_p2m1'_map.
  injection H_p2m1'_map; intro H_p2m1'_eq_p2m1; subst p2m1'; clear H_p2m1'_map.
  destruct H_pre2 as [H_os_running2 [H_trusted_aos2 [[p2m2' H_p2m2'_map H_pa_err2] H_pre2]]].
  rewrite <- H_post_hyp2 in H_p2m2'_map.
  injection H_p2m2'_map; intro H_p2m2'_eq_p2m2; subst p2m2'; clear H_p2m2'_map.

  unfold padd_page.
  rewrite H_s1'; rewrite H_s2'; simpl.
  intros pt_padd pt_page1 mp1 [pt_addr1 H_pt_addr1_map H_pt_page1_map] H_pt_page1_content.

  rewrite H_active_os.

  case (padd_eq pt_padd pa).
  intros H_pt_padd_eq_pa.
  unfold hypervisor_mapping_madd in H_pt_addr1_map.
  destruct H_pt_addr1_map as [p2m1' H_p2m1'_map H_pt_addr1_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intros H_p2m1'_value; subst p2m1'; clear H_p2m1'_map.
  rewrite H_pt_padd_eq_pa in H_pt_addr1_map.
  rewrite lemma_map_add_apply in H_pt_addr1_map.
  injection H_pt_addr1_map; intro H_pt_addr1_eq_ma1; clear H_pt_addr1_map.
  rewrite H_pt_addr1_eq_ma1, lemma_map_add_apply in H_pt_page1_map.
  injection H_pt_page1_map; intro H_pt_page1_value.
  induction type; try (rewrite H_pt_page1_value in H_pt_page1_content; discriminate H_pt_page1_content).

  exists (Page (PT_c (map_empty vadd madd)) (Os_PO (st_active_os s1))).
  exists ma2.
  unfold hypervisor_mapping_madd; rewrite lemma_map_add_apply.
  exists (map_add padd_eq p2m2 pa ma2); auto.
  rewrite H_pt_padd_eq_pa.
  rewrite lemma_map_add_apply; reflexivity.
  rewrite lemma_map_add_apply, H_active_os; reflexivity.
  exists (map_empty vadd madd); auto.

  rewrite H_pt_page1_value in H_pt_page1_content.
  unfold page_content in H_pt_page1_content.
  injection H_pt_page1_content; intro H_mp1_empty.
  subst mp1.
  intros va pa' ma' pg' H_abs.
  simpl in H_abs; discriminate H_abs.
  
  (* If pt_padd is not the pinned page *)
  intro H_pt_padd_ne_pa.
  assert (ma1<>pt_addr1) as H_ma1_ne_pt_addr1.
  intros H_ma1_eq_pt_addr1.
  unfold hypervisor_mapping_madd in H_pt_addr1_map.
  destruct H_pt_addr1_map as [p2m1' H_p2m1'_map H_pt_addr1_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intros; subst p2m1'; clear H_p2m1'_map.
  rewrite <- H_ma1_eq_pt_addr1 in H_pt_addr1_map.
  apply H_pt_padd_ne_pa.
  assert (Value padd ma1 = map_apply padd_eq (map_add padd_eq p2m1 pa ma1) pa) as H_ma1_hyp_map.
  rewrite lemma_map_add_apply; reflexivity.
  rewrite H_pt_addr1_map in H_ma1_hyp_map.

  assert (valid_state ctxt s1') as H_s1'_valid.
  apply valid_state_invariant with s1 (Page_pin_trusted pa type); auto.
  generalize H_s1'_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_valid_currpg H_va_has_pa.
  unfold injective_hyper_mappings in H_injective_hyper.
  rewrite H_injective_hyper with (st_active_os s1) (map_add padd_eq p2m1 pa ma1) pt_padd pa; auto.

  rewrite H_s1'; simpl.
  rewrite lemma_map_add_apply; reflexivity.

  elim H_page_s1s2 with pt_padd pt_page1 mp1.
  intros pt_page2 H_pt_page2_map [mp2 H_pt_page2_content H_virt_map_eq].
  unfold padd_page in H_pt_page2_map.
  destruct H_pt_page2_map as [pt_addr2 H_pt_addr2_map H_pt_page2_map].
  
  assert (ma2<>pt_addr2) as H_ma2_ne_pt_addr2.
  intro H_ma2_eq_pt_addr2.
  rewrite H_ma2_eq_pt_addr2 in H_pg2_map.
  rewrite <- H_pt_page2_map in H_pg2_map.
  injection H_pg2_map; intros; subst pg2.
  rewrite <- H_pt_page2_content in H_pg2_content.
  discriminate H_pg2_content.

  exists pt_page2.
  exists pt_addr2.
  unfold hypervisor_mapping_madd; rewrite lemma_map_add_apply.
  exists (map_add padd_eq p2m2 pa ma2); auto.
  rewrite lemma_map_add_apply_distinct; auto.
  unfold hypervisor_mapping_madd in H_pt_addr2_map.
  destruct H_pt_addr2_map as [p2m2' H_p2m2'_map H_pt_addr2_map].
  rewrite H_active_os in H_p2m2'_map.
  rewrite <- H_post_hyp2 in H_p2m2'_map.
  injection H_p2m2'_map; intros; subst p2m2'; auto.

  rewrite lemma_map_add_apply_distinct; auto.

  exists mp2; auto.
  
  intros va pa' ma' pg' H_ma'_map H_pg'_map H_va_accessible H_pg'_type H_ma'_hyp_map.
  case (padd_eq pa pa').
  intro H_pa_eq_pa'.
  assert (ma1=ma') as H_ma1_eq_ma'.
  unfold hypervisor_mapping_madd in H_ma'_hyp_map.
  destruct H_ma'_hyp_map as [p2m1' H_p2m1'_map H_ma'_hyp_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intros H_p2m1'_value.
  subst p2m1'.
  rewrite <- H_pa_eq_pa' in H_ma'_hyp_map.
  rewrite lemma_map_add_apply in H_ma'_hyp_map.
  injection H_ma'_hyp_map; auto.

  rewrite lemma_map_add_apply_distinct in H_pt_page1_map; auto.
  unfold hypervisor_mapping_madd in H_pt_addr1_map.
  destruct H_pt_addr1_map as [p2m1' H_p2m1'_map H_pt_addr1_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intros; subst p2m1'; clear H_p2m1'_map.
  rewrite lemma_map_add_apply_distinct in H_pt_addr1_map; auto.
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_hyper H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_virtual_mapping in H_valid_virtmap.
  elim H_valid_virtmap with (st_active_os s1) pt_addr1 mp1 pt_page1 va ma'; auto.
  clear H_valid_virtmap.
  intros pg1' H_pg1'_map [H_acc H_not_acc].
  assert (page_owned_by pg1'=Os_PO (st_active_os s1)) as H_pg1'_owner.
  apply H_acc; auto.
  rewrite H_ma1_eq_ma' in H_pg1_map.
  rewrite <- H_pg1_map in H_pg1'_map.
  injection H_pg1'_map; intros; subst pg1'.
  rewrite H_pg1_owner in H_pg1'_owner.
  discriminate H_pg1'_owner.
  clear H_valid_virtmap.
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  elim H_valid_hyper with (st_active_os s1) pt_padd p2m1; auto.
  clear H_valid_hyper; intros H_valid_hyper.
  destruct H_valid_hyper as [pt_addr1' H_pt_addr1'_map H_pt_addr1'_owner].
  unfold maddress_owner in H_pt_addr1'_owner.
  rewrite <- H_pt_addr1_map in H_pt_addr1'_map.
  injection H_pt_addr1'_map; intros; subst pt_addr1'; clear H_pt_addr1'_map.
  rewrite <- H_pt_page1_map in H_pt_addr1'_owner.
  assumption.
  intro H_abs.
  rewrite <- H_pt_addr1_map in H_abs.
  discriminate H_abs.

  intros H_pa_ne_pa'.
  assert (ma1<>ma') as H_ma1_ne_ma'.
  intro H_ma1_eq_ma'.
  apply H_pa_ne_pa'.

  assert (valid_state ctxt s1') as H_s1'_valid.
  apply valid_state_invariant with s1 (Page_pin_trusted pa type); auto.
  generalize H_s1'_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_valid_currpg H_va_has_pa.
  unfold injective_hyper_mappings in H_injective_hyper.
  rewrite H_injective_hyper with (st_active_os s1) (map_add padd_eq p2m1 pa ma1) pa pa'; auto.

  rewrite H_s1'; simpl.
  rewrite lemma_map_add_apply; reflexivity.
  rewrite lemma_map_add_apply.
  unfold hypervisor_mapping_madd in H_ma'_hyp_map.
  destruct H_ma'_hyp_map as [p2m1' H_p2m1'_map H_ma'_hyp_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intro H_p2m1'_value; subst p2m1'; clear H_p2m1'_map.
  rewrite <- H_ma1_eq_ma' in H_ma'_hyp_map.
  assumption.

  elim H_virt_map_eq with va pa' ma' pg'; auto.
  intros ma2' H_ma2'_map [H_ma2'_hyp_map H_pg'_map2].
  exists ma2'; auto.
  destruct H_ma2'_hyp_map as [p2m2' H_p2m2'_map H_ma2'_hyp_map].
  rewrite H_active_os in H_p2m2'_map.
  rewrite <- H_post_hyp2 in H_p2m2'_map.
  injection H_p2m2'_map; intros; subst p2m2'; clear H_p2m2'_map.
  split.
  unfold hypervisor_mapping_madd.
  rewrite lemma_map_add_apply.
  exists (map_add padd_eq p2m2 pa ma2); auto.
  rewrite lemma_map_add_apply_distinct; auto.
  unfold hypervisor_mapping_madd in H_ma2'_hyp_map.

  assert (ma2'<>ma2) as H_ma2'_ne_ma2.
  intro H_ma2'_eq_ma2.
  rewrite H_ma2'_eq_ma2 in H_pg'_map2.
  rewrite <- H_pg2_map in H_pg'_map2.
  injection H_pg'_map2; intro H_pg'_eq_pg2.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_map, H_pg'_type; auto.
  rewrite <- H_pg'_map in H_pg'_type.
  rewrite H_pg'_eq_pg2 in H_pg'_type.
  induction (page_content pg2); discriminate.

  rewrite lemma_map_add_apply_distinct; auto.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  unfold page_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  rewrite <- H_pg'_map.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_type; auto.
  rewrite <- H_pg'_map in H_pg'_type.
  assumption.

  unfold hypervisor_mapping_madd in H_ma'_hyp_map.
  destruct H_ma'_hyp_map as [p2m1' H_p2m1'_map H_ma'_hyp_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intros H_p2m1'_value.
  rewrite H_p2m1'_value in H_ma'_hyp_map.
  rewrite lemma_map_add_apply_distinct in H_ma'_hyp_map; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.

  unfold padd_page.
  exists pt_addr1.
  unfold hypervisor_mapping_madd in H_pt_addr1_map.
  destruct H_pt_addr1_map as [p2m1' H_p2m1'_map H_pt_addr1_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intros H_p2m1'_value.
  rewrite H_p2m1'_value in H_pt_addr1_map.
  rewrite lemma_map_add_apply_distinct in H_pt_addr1_map; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.

  rewrite lemma_map_add_apply_distinct in H_pt_page1_map; auto.
  assumption.
Qed.

Lemma pin_trusted_os_isolation : forall (s1 s2 s1' s2':state) (pa:padd) (type:ptype),
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  active_equivalent ctxt s1 s2 ->
  valid_exec ctxt s1 (Page_pin_trusted pa type) s1' ->
  valid_exec ctxt s2 (Page_pin_trusted pa type) s2' ->
  active_equivalent ctxt s1' s2'.
Proof.
  intros s1 s2 s1' s2' pa type H_s1_valid H_s2_valid H_active_equivalent.
  generalize H_active_equivalent.
  unfold active_equivalent;
    intros (H_active_os, (H_activity_aos,
      (H_oss, ((H_hyp1, H_hyp2), (H_page_s1s2, H_page_s2s1)))), H_exec1, H_exec2).
  generalize H_exec1; generalize H_exec2.
  intros H_exec2' H_exec1'.
  inversion_clear H_exec1'.
  inversion_clear H_exec2'.
  rename H into H_pre1.
  rename H0 into H_post1.
  rename H1 into H_pre2.
  rename H2 into H_post2.
  unfold post_action, post_page_pin_untrusted in H_post1, H_post2.
  unfold post_all_pin in H_post1, H_post2.
  destruct H_post1 as [ma1 [pg1 H_pg1_map [H_pg1_owner
    [H_pg1_content [p2m1 H_post_hyp1 [os1 H_os1_map H_s1']]]]]].
  destruct H_post2 as [ma2 [pg2 H_pg2_map [H_pg2_owner
    [H_pg2_content [p2m2 H_post_hyp2 [os2 H_os2_map H_s2']]]]]].

  generalize (active_eq_symm ctxt s1 s2); intro H_s2_equiv_s1.
  
  assert (st_active_os s1' = st_active_os s2') as H_active_os'.
  rewrite H_s1', H_s2'; simpl; assumption.
  repeat split; try (rewrite H_s1', H_s2'; simpl; assumption).
  
  (* Prove AOS does not change *)
  rewrite H_s1'; rewrite H_s2'; simpl.
  repeat rewrite lemma_map_add_apply.
  assert (os1 = os2).
  rewrite  <- H_oss  in H_os2_map.
  rewrite <- H_os2_map in H_os1_map.
  injection H_os1_map; auto.
  subst os2; reflexivity.

  apply pin_trusted_os_isolation_hyper with s1 s2 pa type; auto.
  rewrite H_active_os'.
  apply pin_trusted_os_isolation_hyper with s2 s1 pa type; auto.
  apply pin_trusted_os_isolation_pt_pages with s1 s2 pa type; auto.
  rewrite H_active_os'.
  apply pin_trusted_os_isolation_pt_pages with s2 s1 pa type; auto.
Qed.

(* Page unpin *)
Lemma unpin_trusted_os_isolation_hyper : forall (s1 s2 s1' s2':state) (pa:padd),
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  active_equivalent ctxt s1 s2 ->
  valid_exec ctxt s1 (Page_unpin_trusted pa) s1' ->
  valid_exec ctxt s2 (Page_unpin_trusted pa) s2' ->
  osi_hyper_mapping_eq s1' s2' (st_active_os s1').
Proof.
  intros s1 s2 s1' s2' pa H_s1_valid H_s2_valid.
  unfold active_equivalent, aos_equivalent_hyper_mapping,
    aos_equivalent_pt_pages, osi_equivalent_hyper_mapping, osi_hyper_mapping_eq;
    intros (H_active_os, (H_activity_aos,
      (H_oss, ((H_hyp1, H_hyp2), (H_page_s1s2, H_page_s2s1)))), H_exec1, H_exec2).
  generalize H_exec1; generalize H_exec2.
  intros H_exec2' H_exec1'.
  inversion_clear H_exec1'.
  inversion_clear H_exec2'.
  rename H into H_pre1.
  rename H0 into H_post1.
  rename H1 into H_pre2.
  rename H2 into H_post2.
  unfold post_action, post_page_unpin_untrusted, post_all_unpin in H_post1, H_post2.
  destruct H_post1 as [p2m1 H_p2m1_map [ma1 H_ma1_hyp_map [system1 H_system1_map H_s1']]].
  destruct H_post2 as [p2m2 H_p2m2_map [ma2 H_ma2_hyp_map [system2 H_system2_map H_s2']]].

  unfold hypervisor_mapping in H_p2m1_map, H_p2m2_map.

  unfold padd_page.
  rewrite H_s1'; rewrite H_s2'; simpl.
  intros pa' pg' H_pg'_map [val H_pg'_content].

  destruct H_pg'_map as [ma' H_ma'_hyp_map H_pg'_map].
  unfold hypervisor_mapping_madd in H_ma'_hyp_map.
  destruct H_ma'_hyp_map as [p2m1' H_p2m1'_map H_ma'_hyp_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intro; subst p2m1'; clear H_p2m1'_map.
  assert (pa<>pa') as H_pa_ne_pa'.
  intro H_pa_eq_pa'.
  rewrite H_pa_eq_pa' in H_ma'_hyp_map.
  rewrite map_drop_apply in H_ma'_hyp_map.
  discriminate H_ma'_hyp_map.
  rewrite map_drop_apply_distinct in H_ma'_hyp_map; auto.

  assert (ma1<>ma') as H_ma1_ne_ma'.
  intro H_ma1_eq_ma'.
  apply H_pa_ne_pa'.
  apply hyp_mapped_same_paddr with ctxt s1 (st_active_os s1) ma'; auto.
  rewrite <- H_ma1_eq_ma'.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  unfold hypervisor_mapping_madd; exists p2m1; auto.

  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.

  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  elim H_valid_hyper with (st_active_os s1) pa' p2m2; auto.
  intros [ma H_ma_hyp_map H_ma_owner].
  clear H_valid_hyper.
  
  exists ma.
  unfold hypervisor_mapping_madd.
  exists (map_drop padd_eq p2m2 pa); auto.
  rewrite H_active_os.
  rewrite lemma_map_add_apply; auto.
  rewrite map_drop_apply_distinct; auto.
  assert (ma <> ma2) as H_ma_ne_ma2.
  intro H_ma_eq_ma2.
  apply H_pa_ne_pa'.
  apply hyp_mapped_same_paddr with ctxt s2 (st_active_os s2) ma2; auto.
  exists p2m2; auto.
  rewrite <- H_active_os.
  rewrite <- H_ma_eq_ma2.
  unfold hypervisor_mapping_madd.
  rewrite H_active_os.
  exists p2m2; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  elim H_hyp1 with pa' pg'.
  intros ma2' H_ma2'_hyp_map H_pg'_map2.
  assert (ma2'=ma).
  rewrite H_active_os in H_ma2'_hyp_map.
  apply hyp_mapped_same_maddr with s2 (st_active_os s2) pa'; auto.
  exists p2m2; auto.
  subst ma2'; clear H_ma2'_hyp_map.
  assumption.

  unfold padd_page.
  exists ma'; auto.
  exists p2m1; auto.
  exists val; auto.
  
  intro H_abs.
  clear H_valid_hyper.
  elim H_hyp1 with pa' pg'.
  intros ma H_ma_hyp_map H_pg'_map2.
  unfold hypervisor_mapping_madd in H_ma_hyp_map.
  destruct H_ma_hyp_map as [p2m2' H_p2m2'_map H_ma_map].
  rewrite <- H_active_os in H_p2m2_map.
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intro; subst p2m2'; clear H_p2m2'_map.
  rewrite <- H_ma_map in H_abs.
  discriminate H_abs.
  unfold padd_page.
  exists ma'; auto.
  exists p2m1; auto.
  exists val; auto.
  rewrite H_active_os; assumption.
Qed.

Lemma unpin_trusted_os_isolation_pt_pages : forall (s1 s2 s1' s2':state) (pa:padd),
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  active_equivalent ctxt s1 s2 ->
  valid_exec ctxt s1 (Page_unpin_trusted pa) s1' ->
  valid_exec ctxt s2 (Page_unpin_trusted pa) s2' ->
  osi_pt_pages_eq ctxt s1' s2' (st_active_os s1').
Proof.
  intros s1 s2 s1' s2' pa H_s1_valid H_s2_valid.
  unfold active_equivalent, aos_equivalent_hyper_mapping,
    aos_equivalent_pt_pages, osi_equivalent_pt_pages, osi_pt_pages_eq;
    intros (H_active_os, (H_activity_aos,
      (H_oss, ((H_hyp1, H_hyp2), (H_page_s1s2, H_page_s2s1)))), H_exec1, H_exec2).
  generalize H_exec1; generalize H_exec2.
  intros H_exec2' H_exec1'.
  inversion_clear H_exec1'.
  inversion_clear H_exec2'.
  rename H into H_pre1.
  rename H0 into H_post1.
  rename H1 into H_pre2.
  rename H2 into H_post2.
  unfold post_action, post_page_unpin_untrusted in H_post1, H_post2.
  unfold post_all_unpin in H_post1, H_post2.
  destruct H_post1 as [p2m1 H_p2m1_map [ma1 H_ma1_hyp_map [system1 H_system1_map H_s1']]].
  destruct H_post2 as [p2m2 H_p2m2_map [ma2 H_ma2_hyp_map [system2 H_system2_map H_s2']]].

  unfold hypervisor_mapping in H_p2m1_map, H_p2m2_map.

  unfold pre_action, pre_page_unpin_untrusted, pre_all_unpin in H_pre1, H_pre2.
  elim H_pre1.
  intros H_aos1_running [H_trusted_aos1
    [system1' H_sys1'_map [H_pa_ne_osi_curr_page1 [
      p2m1' H_p2m1'_map [ma1' H_ma1'_map [H_ma1'_ptype H_no_va_mapped1]]]]]].
  unfold hypervisor_mapping in H_p2m1'_map.
  rewrite <- H_p2m1_map in H_p2m1'_map.
  injection H_p2m1'_map; intro H_p2m1'_eq_p2m1; subst p2m1'; clear H_p2m1'_map.
  elim H_pre2.
  intros H_aos2_running [H_trusted_aos2
    [system2' H_sys2'_map [H_pa_ne_osi_curr_page2 [
      p2m2' H_p2m2'_map [ma2' H_ma2'_map [H_ma2'_ptype H_no_va_mapped2]]]]]].
  unfold hypervisor_mapping in H_p2m2'_map.
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intro H_p2m2'_eq_p2m2; subst p2m2'; clear H_p2m2'_map.
  rewrite <- H_system1_map in H_sys1'_map.
  injection H_sys1'_map; intro; subst system1'; clear H_sys1'_map.
  rewrite <- H_system2_map in H_sys2'_map.
  injection H_sys2'_map; intro; subst system2'; clear H_sys2'_map.
  rewrite <- H_ma1_hyp_map in H_ma1'_map.
  injection H_ma1'_map; intro; subst ma1'; clear H_ma1'_map.
  rewrite <- H_ma2_hyp_map in H_ma2'_map.
  injection H_ma2'_map; intro; subst ma2'; clear H_ma2'_map.
  
  unfold padd_page.
  rewrite H_s1'; rewrite H_s2'; simpl.
  intros pt_padd pt_page1 mp1 [pt_addr1 H_pt_addr1_map H_pt_page1_map] H_pt_page1_content.

  rewrite H_active_os.

  (* pt_padd cannot be the unpinned page *)
  assert (pt_padd<>pa) as H_pt_padd_ne_pa.
  intros H_pt_padd_eq_pa.
  unfold hypervisor_mapping_madd in H_pt_addr1_map.
  destruct H_pt_addr1_map as [p2m1' H_p2m1'_map H_pt_addr1_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intros H_p2m1'_value; subst p2m1'; clear H_p2m1'_map.
  rewrite H_pt_padd_eq_pa in H_pt_addr1_map.
  rewrite map_drop_apply in H_pt_addr1_map.
  discriminate H_pt_addr1_map.
  
  unfold hypervisor_mapping_madd in H_pt_addr1_map.
  destruct H_pt_addr1_map as [p2m1' H_p2m1'_map H_pt_addr1_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intros; subst p2m1'; clear H_p2m1'_map.
  rewrite map_drop_apply_distinct in H_pt_addr1_map; auto.

  assert (ma1<>pt_addr1) as H_ma1_ne_pt_addr1.
  intros H_ma1_eq_pt_addr1.
  rewrite <- H_ma1_eq_pt_addr1 in H_pt_addr1_map.
  apply H_pt_padd_ne_pa.
  rewrite H_ma1_hyp_map in H_pt_addr1_map.

  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_valid_currpg H_va_has_pa.
  unfold injective_hyper_mappings in H_injective_hyper.
  rewrite H_injective_hyper with (st_active_os s1) p2m1 pt_padd pa; auto.

  elim H_page_s1s2 with pt_padd pt_page1 mp1; auto.
  intros pt_page2 H_pt_page2_map [mp2 H_pt_page2_content H_virt_map_eq].
  unfold padd_page in H_pt_page2_map.
  destruct H_pt_page2_map as [pt_addr2 H_pt_addr2_map H_pt_page2_map].
  unfold hypervisor_mapping_madd in H_pt_addr2_map.
  destruct H_pt_addr2_map as [p2m2' H_p2m2'_map H_pt_addr2_map].
  rewrite <- H_active_os in H_p2m2_map.
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intros; subst p2m2'; clear H_p2m2'_map.

  assert (ma2<>pt_addr2) as H_ma2_ne_pt_addr2.
  intro H_ma2_eq_pt_addr2.
  rewrite <- H_ma2_eq_pt_addr2 in H_pt_addr2_map.
  apply H_pt_padd_ne_pa.
  rewrite H_ma2_hyp_map in H_pt_addr2_map.

  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_valid_currpg H_va_has_pa.
  unfold injective_hyper_mappings in H_injective_hyper.
  rewrite H_injective_hyper with (st_active_os s2) p2m2 pt_padd pa; auto.
  rewrite <- H_active_os; auto.

  exists pt_page2.
  exists pt_addr2.
  unfold hypervisor_mapping_madd; rewrite lemma_map_add_apply.
  exists (map_drop padd_eq p2m2 pa); auto.
  rewrite map_drop_apply_distinct; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists mp2; auto.
  
  intros va pa' ma' pg' H_ma'_map H_pg'_map H_va_accessible H_pg'_type H_ma'_hyp_map.

  unfold hypervisor_mapping_madd in H_ma'_hyp_map.
  destruct H_ma'_hyp_map as [p2m1' H_p2m1'_map H_ma'_hyp_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intro H_p2m1'_value.
  rewrite H_p2m1'_value in H_ma'_hyp_map.
  assert (pa<>pa') as H_pa_ne_pa'.
  intro H_pa_eq_pa'.
  rewrite H_pa_eq_pa' in H_ma'_hyp_map.
  rewrite map_drop_apply in H_ma'_hyp_map.
  discriminate H_ma'_hyp_map.

  assert (ma1<>ma') as H_ma1_ne_ma'.
  intro H_ma1_eq_ma'.
  rewrite map_drop_apply_distinct in H_ma'_hyp_map; auto.
  apply H_pa_ne_pa'.

  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_valid_currpg H_va_has_pa.
  unfold injective_hyper_mappings in H_injective_hyper.
  rewrite H_injective_hyper with (st_active_os s1) p2m1 pa pa'; auto.

  rewrite <- H_ma'_hyp_map.
  rewrite <- H_ma1_eq_ma'; auto.
  
  rewrite lemma_map_add_apply_distinct in H_pt_page1_map; auto.
  rewrite map_drop_apply_distinct in H_ma'_hyp_map; auto.

  elim H_virt_map_eq with va pa' ma' pg'; auto.
  intros ma2' H_ma2'_map [H_ma2'_hyp_map H_pg'_map2].
  exists ma2'; auto.
  destruct H_ma2'_hyp_map as [p2m2' H_p2m2'_map H_ma2'_hyp_map].
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intros; subst p2m2'; clear H_p2m2'_map.
  split.
  unfold hypervisor_mapping_madd.
  rewrite lemma_map_add_apply.
  exists (map_drop padd_eq p2m2 pa); auto.
  rewrite map_drop_apply_distinct; auto.

  assert (ma2'<>ma2) as H_ma2'_ne_ma2.
  intro H_ma2'_eq_ma2.
  rewrite H_ma2'_eq_ma2 in H_pg'_map2.
  apply H_pa_ne_pa'.

  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_valid_currpg H_va_has_pa.
  unfold injective_hyper_mappings in H_injective_hyper.
  rewrite H_injective_hyper with (st_active_os s2) p2m2 pa pa'; auto.

  rewrite <- H_active_os; auto.
  
  rewrite <- H_ma2'_hyp_map.
  rewrite H_ma2'_eq_ma2; auto.

  rewrite lemma_map_add_apply_distinct; auto.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.

  unfold page_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  rewrite <- H_pg'_map.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_type; auto.
  rewrite <- H_pg'_map in H_pg'_type.
  assumption.

  unfold hypervisor_mapping_madd.
  exists p2m1; auto.

  unfold padd_page.
  exists pt_addr1.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.

  rewrite lemma_map_add_apply_distinct in H_pt_page1_map; auto.
Qed.

Lemma unpin_trusted_os_isolation : forall (s1 s2 s1' s2':state) (pa:padd),
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  active_equivalent ctxt s1 s2 ->
  valid_exec ctxt s1 (Page_unpin_trusted pa) s1' ->
  valid_exec ctxt s2 (Page_unpin_trusted pa) s2' ->
  active_equivalent ctxt s1' s2'.
Proof.
  intros s1 s2 s1' s2' pa H_s1_valid H_s2_valid H_active_equivalent.
  generalize H_active_equivalent.
  unfold active_equivalent;
    intros (H_active_os, (H_activity_aos,
      (H_oss, ((H_hyp1, H_hyp2), (H_page_s1s2, H_page_s2s1)))), H_exec1, H_exec2).
  generalize H_exec1; generalize H_exec2.
  intros H_exec2' H_exec1'.
  inversion_clear H_exec1'.
  inversion_clear H_exec2'.
  rename H into H_pre1.
  rename H0 into H_post1.
  rename H1 into H_pre2.
  rename H2 into H_post2.
  unfold post_action, post_page_unpin_untrusted in H_post1, H_post2.
  unfold post_all_unpin in H_post1, H_post2.
  destruct H_post1 as [p2m1 H_p2m1_map [ma1 H_ma1_hyp_map [system1 H_system1_map H_s1']]].
  destruct H_post2 as [p2m2 H_p2m2_map [ma2 H_ma2_hyp_map [system2 H_system2_map H_s2']]].

  unfold hypervisor_mapping in H_p2m1_map, H_p2m2_map.

  generalize (active_eq_symm ctxt s1 s2); intro H_s2_equiv_s1.
  
  assert (st_active_os s1' = st_active_os s2') as H_active_os'.
  rewrite H_s1'; rewrite H_s2'; assumption.
  repeat split; try (rewrite H_s1'; rewrite H_s2'; simpl; assumption).
  
  (* Prove AOS does not change *)
  rewrite H_s1'; rewrite H_s2'; simpl.
  rewrite <- H_active_os.
  repeat rewrite lemma_map_add_apply.
  assert (system1 = system2).
  rewrite <- H_oss  in H_system2_map.
  rewrite <- H_system2_map in H_system1_map.
  injection H_system1_map; auto.
  subst system2; reflexivity.

  intros H_osi_ne_aos.
  repeat rewrite lemma_map_add_apply_distinct; auto.

  apply unpin_trusted_os_isolation_hyper with s1 s2 pa; auto.
  rewrite H_active_os'.
  apply unpin_trusted_os_isolation_hyper with s2 s1 pa; auto.
  apply unpin_trusted_os_isolation_pt_pages with s1 s2 pa; auto.
  rewrite H_active_os'.
  apply unpin_trusted_os_isolation_pt_pages with s2 s1 pa; auto.
Qed.


(** An OS operation does not depend on other OSs state. *)
Theorem os_isolation :
  forall (s1 s1' s2 s2':state) (a:action),
     valid_state ctxt s1 ->
     valid_state ctxt s2 ->
     active_equivalent ctxt s1 s2 ->
     valid_exec ctxt s1 a s1' ->
     valid_exec ctxt s2 a s2' ->
     (forall osi, a <> Switch osi) -> 
     active_equivalent ctxt s1' s2'.
Proof.
intros s1 s1' s2 s2' a H_s1_valid H_s2_valid.
intro H_active_eq; generalize H_active_eq; unfold active_equivalent;
  intros (H_active_os, (H_activity_aos, (H_oss, (H_page_s1s2, H_page_s2s1)))).  
induction a; simpl; intros valid_exec_s1_s1' valid_exec_s2_s2' condition; generalize valid_exec_s1_s1'; generalize valid_exec_s2_s2'; 
inversion_clear valid_exec_s1_s1'; inversion_clear valid_exec_s2_s2';
  unfold pre_action in H; unfold pre_action in H1; unfold post_action in H0;
    unfold post_action in H2; intros valid_exec_s1_s1' valid_exec_s2_s2'.

(* Silent action (state does not change). *)
inversion H0; inversion H2.         
rewrite <- H3; rewrite <- H4; auto. 

(* Read action (state does not change) *)
inversion H0; inversion H2.     
rewrite <- H3; rewrite <- H4; auto.

(* Read hyper action (state does not change) *)
inversion H0; inversion H2.     
rewrite <- H3; rewrite <- H4; auto.

(* Write action *)
apply write_os_isolation with s1 s2 v v0; auto.
(* Write hyper action *)
apply write_os_isolation with s1 s2 v v0; auto.

(* New untrusted action *)
generalize osi_aos_equivalent.
intros H_osi_aos.
unfold active_equivalent in H_osi_aos.
apply H_osi_aos with (st_active_os s1'); auto.
apply new_os_isolation with s1 s2 o v p; auto.
apply aos_osi_equivalent; auto.
unfold post_new_untrusted, post_all_new in H0.
elim H0.
intros ma H_ma_map [pt_addr H_curr_pt_page [newmem H_newmem [system H_system_map H_assign]]].
unfold state_components_assign, state_components_update in H_assign.
elim H_assign with Active_OS.
intros [H_triv H_interp].
unfold state_component_interp in H_interp; auto.
intros [H_abs H_interp]; contradiction H_abs.
split; intro H_abs'; discriminate H_abs'.


(* New trusted action *)
generalize osi_aos_equivalent.
intros H_osi_aos.
unfold active_equivalent in H_osi_aos.
apply H_osi_aos with (st_active_os s1'); auto.
apply new_os_isolation with s1 s2 (st_active_os s1) v p; auto.
apply aos_osi_equivalent; auto.
unfold post_new_untrusted, post_all_new in H0.
elim H0.
intros ma H_ma_map [pt_addr H_curr_pt_page [newmem H_newmem [system H_system_map H_assign]]].
unfold state_components_assign, state_components_update in H_assign.
elim H_assign with Active_OS.
intros [H_triv H_interp].
unfold state_component_interp in H_interp; auto.
intros [H_abs H_interp]; contradiction H_abs.
split; intro H_abs'; discriminate H_abs'.
unfold post_new_trusted in H2; rewrite H_active_os; auto.
(* New hyper action *)
apply new_hyper_os_isolation with s1 s2 v m; auto.

(* Del untrusted action *)
generalize osi_aos_equivalent.
intros H_osi_aos.
unfold active_equivalent in H_osi_aos.
apply H_osi_aos with (st_active_os s1'); auto.
apply del_untrusted_os_isolation with s1 s2 o v; auto.
apply aos_osi_equivalent; auto.
unfold post_del_untrusted in H0.
destruct H0 as [system H_system_map [p2m H_p2m_map [pt_addr H_pt_addr_map [newmem H_newmem H_assign]]]].
unfold state_components_assign, state_components_update in H_assign.
elim H_assign with Active_OS.
intros [H_triv H_interp].
unfold state_component_interp in H_interp; auto.
intros [H_abs H_interp]; contradiction H_abs.
split; intro H_abs'; discriminate H_abs'.
(* Del trusted action *)
unfold post_del_trusted in H0, H2.
destruct H0 as [pt_addr1 H_pt_addr1_map H_post1].
destruct H2 as [pt_addr2 H_pt_addr2_map H_post2].
apply del_os_isolation with s1 s2 v pt_addr1 pt_addr2; auto.
(* Del hyper action *)
unfold post_del_trusted in H0, H2.
destruct H0 as [pt_addr1 H_pt_addr1_map H_post1].
destruct H2 as [pt_addr2 H_pt_addr2_map H_post2].
apply del_os_isolation with s1 s2 v pt_addr1 pt_addr2; auto.

(* Switch action *)
cut False; [intro FALSE; elim FALSE | apply (condition o); trivial].

(* Local Switch trusted action *)
apply lswitch_os_isolation with s1 s2 (st_active_os s1) p;
  try (rewrite H_active_os; assumption); assumption.
(* Local Switch untrusted action *)
apply lswitch_os_isolation with s1 s2 o p; assumption.

(* Hcall action *)
apply hcall_os_isolation with s1 s2 h; assumption.

(* Return Control action *)
apply ret_ctrl_os_isolation with s1 s2; assumption.

(* Chmod action *)
apply chmod_os_isolation with s1 s2; assumption.

(* Page pin/unpin trusted *)
apply pin_trusted_os_isolation with s1 s2 p p0; assumption.
apply unpin_trusted_os_isolation with s1 s2 p; assumption.

(* Page pin untrusted *)
generalize osi_aos_equivalent.
intros H_osi_aos.
unfold active_equivalent in H_osi_aos.
apply H_osi_aos with (st_active_os s1'); auto.
apply pin_untrusted_os_isolation with s1 s2 o p p0; auto.
apply aos_osi_equivalent; auto.
unfold post_page_pin_untrusted, post_all_pin in H0.
destruct H0 as [ma [pg H_pg_map [H_pg_owner [H_pg_content [p2m h_p2m_map [system H_system_map H_s1']]]]]].
rewrite H_s1'; simpl; auto.

(* Page unpin untrusted *)
generalize osi_aos_equivalent.
intros H_osi_aos.
unfold active_equivalent in H_osi_aos.
apply H_osi_aos with (st_active_os s1'); auto.
apply unpin_untrusted_os_isolation with s1 s2 o p; auto.
apply aos_osi_equivalent; auto.
unfold post_page_unpin_untrusted, post_all_unpin in H0.
destruct H0 as [p2m H_p2m_map [ma H_ma_map [system H_system_map H_s1']]].
rewrite H_s1'; simpl; auto.
Qed.


End Full_OS_Isolation.
