(***********************************************************************
 * File: Maps_proof.v
 * $Revision: 408 $
 * $Date: 2011-01-03 13:23:41 -0200 (Mon, 03 Jan 2011) $
 * $LastChangedBy: jdcampo $
 * ********************************************************************
 * 
 ************************************************************************)

Require Export Maps.

Set Implicit Arguments.

(* This module describes properties (and eventually proofs) of mappings *)

Section Mapping_Properties.

(* The type of index for accessing the map *)
Variable index : Set.
Variable index_eq : forall x y : index, {x = y} + {x <> y}.

(* The type of information associated to each index *)
Variable info : Set.

Definition injective_mapping 
  (mp:mapping index info) : Prop :=
  forall id1,
   forall id2,
      map_apply index_eq mp id1 = map_apply index_eq mp id2 -> id1 = id2.  


(* Obtaining an element in index I, after having added VAL in I, gives
   VAL as result *)
Lemma lemma_map_add_apply :
  forall (map:mapping index info) (i:index) (val:info),
    map_apply index_eq (map_add index_eq map i val) i = Value index val.
Proof.
  intros map i val.
  induction map.
  simpl.
  elim (index_eq i i).
  intro; reflexivity.
  intro H_i_ne_i; cut False; (contradiction || apply H_i_ne_i; reflexivity).

  simpl.
  case (index_eq (item_index a) i).
  intro H_aindex_eq_i.
  simpl.
  elim (index_eq i i).
  intro; reflexivity.

  intro H_i_ne_i; cut False; (contradiction || apply H_i_ne_i; reflexivity).

  intro H_aindex_ne_i.
  simpl.
  elim (index_eq i (item_index a)).
  intro H_abs; cut False.
  intro; contradiction.
  apply H_aindex_ne_i; symmetry; assumption.


  intro; assumption.
Qed.

(* Obtaining an element in index I, after having added VAL in I', does
   not change the result. *)
Lemma lemma_map_add_apply_distinct :
  forall (map: mapping index info) (i i':index) (val:info),
    i <> i' -> 
    map_apply index_eq (map_add index_eq map i val) i' = 
    map_apply index_eq map i'.
Proof.
  intros map i i' val H_i_ne_i'.
  induction map.

  simpl.
  elim (index_eq i' i).
  intro H_abs; cut False; (contradiction || auto).
  auto.

  simpl.
  case (index_eq (item_index a) i).
  intro H_aindex_eq_i.
  simpl.
  elim (index_eq i' i).
  intro H_asb; cut False; (contradiction || auto).
  intro H; clear H.
  
  case (index_eq i' (item_index a)).
  intros H_abs.
  rewrite H_aindex_eq_i in H_abs.
  symmetry in H_abs; cut False; (contradiction || auto).
  intro H_i'_ne_aindex.
  reflexivity.

  intro H_aindex_ne_i.
  simpl.
  case (index_eq i' (item_index a)).
  intro H_i'_eq_aindex.
  reflexivity.

  intro H_i'_ne_aindex.
  assumption.
Qed.

Lemma map_drop_apply : forall (map:mapping index info) (i:index),
    map_apply index_eq (map_drop index_eq map i) i = Error info i.
Proof.
  intros map i.
  induction map.
  trivial.
  simpl.

  destruct (index_eq i (item_index a));auto.
  simpl.
  destruct (index_eq i (item_index a)). 
  contradiction.
  assumption.
Qed.

Lemma map_drop_apply_distinct : forall (map:mapping index info) (i i':index),
  i <> i' -> map_apply index_eq (map_drop index_eq map i) i' = map_apply index_eq map i'.
Proof.
  intros map i i' H_i_ne_i'.
  induction map.

  trivial.

  simpl map_drop.
  case (index_eq i (item_index a)).
  intro H_i_index_a.
  destruct (index_eq (item_index a) i').
  rewrite <- H_i_index_a in e.
  contradiction.
  rewrite IHmap.
  simpl.
  destruct (index_eq i' (item_index a)).
  cut False; (contradiction || auto).

  trivial.
  
  intro H_i_ne_index_a.
  simpl.
  destruct (index_eq i' (item_index a)).
  trivial.
  assumption.
Qed.

Lemma map_error_index_inv : forall (map:mapping index info) (i i':index),
  i<>i'->  map_apply index_eq map i <> Error info i'.
Proof.
  intros map i i' H_i_ne_i'.

  induction map.

  intro H_abs;inversion H_abs;auto.

  unfold map_apply.
  case (index_eq i (item_index a)).
  intro H_i_a.
  intro H_abs;inversion H_abs.

  intro H_no_i_a.
  assumption.
Qed.


End Mapping_Properties.
