(***********************************************************************
 * File: Memory.v
 * $Revision: 364 $
 * $Date: 2010-11-04 14:16:55 -0200 (Thu, 04 Nov 2010) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * This file specifies the Memory
 * 
 *************************************************************************)

Require Export Maps.

(** *** Memory and adresses *)

(** Virtual addresses. *)
Parameter vadd: Set.
Parameter vadd_eq : forall va1 va2 : vadd, {va1 = va2} + {va1 <> va2}.

(** Machine addresses. *)
Parameter madd :  Set.
Parameter madd_eq : forall ma1 ma2 : madd, {ma1 = ma2} + {ma1 <> ma2}.

(** Physical addresses : 
The OSes use the physical addresses in order to see 
contiguous regions of memory. They don't see the machine addresses. *)
Parameter padd: Set.
Parameter padd_eq : forall pa1 pa2 : padd, {pa1 = pa2} + {pa1 <> pa2}.

(** Memory values. *)
Parameter value: Set.
Parameter value_eq:forall val1 val2 : value, {val1 = val2} + {val1 <> val2}.

(** We consider that each address that appears in a memory mapping 
    corresponds to a memory page. 
    Each page frame is associated to a type ([ptype])
    that can be : page table (PT), or writable (RW). 
*)

Inductive ptype : Set :=
(** Page Table PD/PT *)
| PT : ptype 
(** Read Write Access *)
| RW: ptype.

(* If madd is a page table then content = vadd -> option madd.
   If madd is a RW table then content = option value *)
Inductive content : Set :=
  (** writable (RW) *)
  | RW_c : option value -> content
  (** page table *)
  (* | PT_c : (vadd -> option madd) -> content *)
  | PT_c : mapping vadd madd -> content 
  (** Neither a RW nor a PT *)
  | Other_c : content. 
