(***********************************************************************
 * File: Read_Isolation.v
 * $Revision: 399 $
 * $Date: 2010-12-26 17:56:13 -0200 (Sun, 26 Dec 2010) $
 * $LastChangedBy: cluna $
 ***********************************************************************
 * This files proves read isolation property on valid executions. 
 **********************************************************************)

Require Import State_valid.
Require Import Env.
Require Import Exec.
Require Import Exec_invariant.
Require Export Observers.

Section Read_Isolation.

Variable ctxt : context.

(** If a read action is executed in a valid state, 
    then the vadd being read has a madd ma associated 
    in the active memory mapping and ma belongs to the Active OS 
*)

Definition read_isolation (s s':state) : Prop :=
  forall (va:vadd), 
         valid_state ctxt s -> 
         valid_exec ctxt s (Read va) s' ->
 	(exists2 ma:madd, va_mapped_to_ma s (st_active_os s) va ma &
		        exists2 pg, Value _ pg = map_apply madd_eq (st_memory s) ma & 
		        Os_PO (st_active_os s) = page_owned_by pg).

Theorem read_isolation_proof : forall (s s':state), read_isolation s s'.
Proof.
intros s s' va.
Intros_Valid_Hyp.
intro valid_read.
inversion_clear valid_read.
inversion_clear H; clear H0.
unfold pre_all_read in H1. 
elim H1; clear H1; intros ma (va_ma, ma_ptype).
unfold valid_virtual_mapping in H_valid_virtmap.
exists ma; trivial.
unfold va_mapped_to_ma in va_ma.
elim va_ma; clear va_ma; intros ma' current_vadd_ma' mp_va_ma.
elim mp_va_ma; clear mp_va_ma; intros mp_va_ma vadd_map_ma'_mp_va_ma ma_mp_va_ma.
unfold vadd_mapping in vadd_map_ma'_mp_va_ma.
destruct vadd_map_ma'_mp_va_ma as (pg, H_pg_map, H_pg_content).
elim H_valid_virtmap with (osi:=(st_active_os s)) (pt_addr:=ma') (mp:=mp_va_ma) (pt:=pg) (va:=va) (ma:=ma);
  clear H_valid_virtmap; intros; trivial. 
exists x; trivial.
symmetry; elim H0; intros.
elim H2; auto.
unfold os_current_vadd_mapping_madd in current_vadd_ma'.
destruct current_vadd_ma' as (aos, H_aos_map, (p2m, H_p2m_map, H_ma'_map)).
unfold valid_hypervisor in H_valid_hyper.
destruct H_valid_hyper with (st_active_os s) (os_current_page aos) p2m.
unfold hypervisor_mapping in H_p2m_map; assumption.

destruct H as (ma'', H_ma''_map, H_ma''_owner).
assert (ma' = ma'').
rewrite <- H_ma''_map in H_ma'_map.
injection H_ma'_map; auto.
subst ma''.
unfold maddress_owner in H_ma''_owner.
rewrite <- H_pg_map in H_ma''_owner.
assumption.

cut False; [intro; contradiction | rewrite H in H_ma'_map].
discriminate H_ma'_map.
Qed.

End Read_Isolation.